from __future__ import annotations
from sqloquent import HashedModel, RelatedModel, RelatedCollection, Default
from sqloquent.interfaces import QueryBuilderProtocol
from tapescript import run_auth_scripts, Script
from .AccountType import AccountType
from .Entry import Entry
from .EntryType import EntryType
import packify


class Account(HashedModel):
    connection_info: str = ''
    table: str = 'accounts'
    id_column: str = 'id'
    columns: tuple[str] = (
        'id', 'name', 'type', 'ledger_id', 'parent_id', 'code',
        'locking_scripts', 'category_id', 'details', 'active'
    )
    columns_excluded_from_hash = ('active',)
    id: str
    name: str
    type: str
    ledger_id: str
    parent_id: str
    code: str|None
    locking_scripts: bytes|None
    category_id: str|None
    details: bytes|None
    active: bool|Default[True]
    ledger: RelatedModel
    parent: RelatedModel
    category: RelatedModel
    children: RelatedCollection
    entries: RelatedCollection
    archived_entries: RelatedCollection

    # override automatic property
    @property
    def type(self) -> AccountType:
        """The AccountType of the Account."""
        return AccountType(self.data['type'])
    @type.setter
    def type(self, val: AccountType):
        if type(val) is AccountType:
            self.data['type'] = val.value

    # override automatic property
    @property
    def locking_scripts(self) -> dict[EntryType, bytes]:
        """The dict mapping EntryType to tapescript locking script bytes."""
        return {
            EntryType(k): v
            for k,v in packify.unpack(
                self.data.get('locking_scripts', None) or b'M@\x00'
            ).items()
        }
    @locking_scripts.setter
    def locking_scripts(self, vals: dict[EntryType, bytes]):
        if type(vals) is dict and all([type(k) is EntryType for k, _ in vals.items()]) \
            and all([type(v) is bytes for k,v in vals.items()]):
            self.data['locking_scripts'] = packify.pack({
                k.value: v
                for k,v in vals.items()
            })

    # override automatic property
    @property
    def details(self) -> packify.SerializableType:
        """A packify.SerializableType stored in the database as a blob."""
        return packify.unpack(self.data.get('details', None) or b'\x00')
    @details.setter
    def details(self, val: packify.SerializableType):
        if isinstance(val, packify.SerializableType):
            self.data['details'] = packify.pack(val)

    @staticmethod
    def _encode(data: dict|None) -> dict|None:
        """Encode Account data without modifying the original dict."""
        if type(data) is not dict:
            return data
        data = {**data}
        if type(data.get('type', None)) is AccountType:
            data['type'] = data['type'].value
        return data

    @classmethod
    def insert(cls, data: dict) -> Account | None:
        """Ensure data is encoded before inserting."""
        result = super().insert(cls._encode(data))
        return result

    @classmethod
    def insert_many(cls, items: list[dict], /, *, suppress_events: bool = False) -> int:
        """Ensure items are encoded before inserting."""
        items = [cls._encode(item) for item in items]
        return super().insert_many(items, suppress_events=suppress_events)

    def update(self, updates: dict, /, *, suppress_events: bool = False) -> Account:
        """Ensure updates are encoded before updating."""
        updates = self._encode(updates)
        return super().update(updates, suppress_events=suppress_events)

    @classmethod
    def query(cls, conditions: dict = None, connection_info: str = None) -> QueryBuilderProtocol:
        """Ensure conditions are encoded before querying."""
        if conditions and type(conditions.get('type', None)) is AccountType:
            conditions['type'] = conditions['type'].value
        return super().query(conditions, connection_info)

    def balance(self, include_sub_accounts: bool = True,
                rolled_up_balances: dict[str, tuple[EntryType, int]] = {}) -> int:
        """Tally all entries for this account. Includes the balances of
            all sub-accounts if include_sub_accounts is True. To get an
            accurate balance, pass in the balances from the most recent
            TxRollup.
        """
        totals = {
            EntryType.CREDIT: 0,
            EntryType.DEBIT: 0,
            'subaccounts': 0,
        }
        if self.id in rolled_up_balances:
            if rolled_up_balances[self.id][0] == EntryType.CREDIT:
                totals[EntryType.CREDIT] = rolled_up_balances[self.id][1]
            else:
                totals[EntryType.DEBIT] = rolled_up_balances[self.id][1]

        for entries in self.entries().query().chunk(500):
            entry: Entry
            for entry in entries:
                totals[entry.type] += entry.amount

        if include_sub_accounts:
            for acct in self.children:
                acct: Account
                totals['subaccounts'] += acct.balance(
                    include_sub_accounts=True,
                    rolled_up_balances=rolled_up_balances
                )

        if self.type in (
            AccountType.ASSET, AccountType.DEBIT_BALANCE,
            AccountType.CONTRA_LIABILITY, AccountType.CONTRA_EQUITY,
            AccountType.NOSTRO_ASSET
        ):
            return totals[EntryType.DEBIT] - totals[EntryType.CREDIT] + totals['subaccounts']

        return totals[EntryType.CREDIT] - totals[EntryType.DEBIT] + totals['subaccounts']

    def validate_script(self, entry_type: EntryType, auth_script: bytes|Script,
                        tapescript_runtime: dict = {}) -> bool:
        """Checks if the auth_script validates against the correct
            locking_script for the EntryType. Returns True if it does
            and False if it does not (or if it errors).
        """
        cache = tapescript_runtime.get('cache', {})
        contracts = tapescript_runtime.get('contracts', {})
        locking_script = self.locking_scripts.get(entry_type, b'')
        if type(auth_script) is Script:
            auth_script = auth_script.bytes
        return run_auth_scripts([auth_script, locking_script], cache, contracts)



from __future__ import annotations
from sqloquent import HashedModel, RelatedCollection
from sqloquent.errors import vert, tert
from .Account import Account, AccountType
from .Correspondence import Correspondence
from .ArchivedEntry import ArchivedEntry, EntryType
from .Identity import Identity
import packify


class ArchivedTransaction(HashedModel):
    """Optional class for storing a trimmed Transaction after is has
        included in a TxRollup. This allows accessing the trimmed
        Transaction details more efficiently than by loading the
        DeletedModel that contains the trimmed Transaction. Must be used
        in conjunction with ArchivedEntry.
    """
    connection_info: str = ''
    table: str = 'archived_transactions'
    id_column: str = 'id'
    columns: tuple[str] = ('id', 'entry_ids', 'ledger_ids', 'timestamp', 'details', 'auth_scripts')
    columns_excluded_from_hash: tuple[str] = ('auth_scripts',)
    id: str
    entry_ids: str
    ledger_ids: str
    timestamp: str
    details: bytes
    auth_scripts: bytes
    entries: RelatedCollection
    ledgers: RelatedCollection

    # override automatic properties
    @property
    def details(self) -> dict[str, bytes]:
        """A packify.SerializableType stored in the database as a blob."""
        return packify.unpack(self.data.get('details', b'M@\x00'))
    @details.setter
    def details(self, val: dict[str, bytes]):
        if type(val) is not dict:
            return
        if not all([type(k) is str and type(v) is bytes for k, v in val.items()]):
            return
        self.data['details'] = packify.pack(val)

    @property
    def auth_scripts(self) -> dict[str, bytes]:
        """A dict mapping account IDs to tapescript unlocking script bytes."""
        return packify.unpack(self.data.get('auth_scripts', b'M@\x00'))
    @auth_scripts.setter
    def auth_scripts(self, val: dict[str, bytes]):
        if type(val) is not dict:
            return
        if not all([type(k) is str and type(v) is bytes for k, v in val.items()]):
            return
        self.data['auth_scripts'] = packify.pack(val)

    @classmethod
    def _encode(cls, data: dict|None) -> dict|None:
        """Encode values for saving."""
        if type(data) is not dict:
            return None
        if type(data.get('auth_scripts', {})) is dict:
            data['auth_scripts'] = packify.pack(data.get('auth_scripts', {}))
        if type(data.get('details', {})) is dict:
            data['details'] = packify.pack(data.get('details', {}))
        return data

    def validate(self, tapescript_runtime: dict = {}, reload: bool = False) -> bool:
        """Determines if a Transaction is valid using the rules of accounting
            and checking all auth scripts against their locking scripts. The
            tapescript_runtime can be scoped to each entry ID. Raises TypeError
            for invalid arguments. Raises ValueError if the entries do not
            balance for each ledger; if a required auth script is missing; or
            if any of the entries is contained within an existing Transaction.
            If reload is set to True, entries and accounts will be reloaded
            from the database.
        """
        tert(type(self.auth_scripts) is dict,
            'auth_scripts must be dict mapping account ID to authorizing tapescript bytecode')

        if reload:
            self.entries().reload()

        # first check that all ledgers balance
        ledgers = {}
        entry: ArchivedEntry
        for entry in self.entries:
            vert(entry.account_id in self.auth_scripts or not entry.account.locking_scripts
                 or entry.type not in entry.account.locking_scripts,
                f"missing auth script for account {entry.account_id} ({entry.account.name})")
            if reload:
                entry.account().reload()
            if entry.account.ledger_id not in ledgers:
                ledgers[entry.account.ledger_id] = {'Dr': 0, 'Cr': 0}
            if entry.type in (EntryType.CREDIT, EntryType.CREDIT.value):
                ledgers[entry.account.ledger_id]['Cr'] += entry.amount
            else:
                ledgers[entry.account.ledger_id]['Dr'] += entry.amount

        for ledger_id, balances in ledgers.items():
            vert(balances['Cr'] == balances['Dr'],
                f"ledger {ledger_id} unbalanced: {balances['Cr']} Cr != {balances['Dr']} Dr")

        # next check that all necessary authorizations are provided
        for entry in self.entries:
            acct = entry.account

            if not acct.locking_scripts or entry.type not in acct.locking_scripts:
                continue
            if acct.id not in self.auth_scripts:
                return False
            runtime = tapescript_runtime.get(entry.id, {**tapescript_runtime})
            if 'cache' not in runtime:
                runtime['cache'] = {}
            if 'sigfield1' not in runtime['cache']:
                runtime['cache'] = {
                    **runtime['cache'],
                    **entry.get_sigfields(tapescript_runtime=tapescript_runtime)
                }
            if not acct.validate_script(entry.type, self.auth_scripts[acct.id], runtime):
                return False

        # finally check that correspondent accounting is not violated
        if len(ledgers) > 1:
            accounts = [e.account for e in self.entries]

            for acct in accounts:
                acct: Account
                if acct.type is AccountType.NOSTRO_ASSET and type(acct.details) is str \
                    and acct.details != acct.id and Identity.find(acct.details):
                    # Nostro account must have equivalent Vostro account
                    nostro = acct
                    nostro.ledger().reload()
                    cor: Correspondence = Correspondence.query().contains(
                        'identity_ids', nostro.ledger.identity_id
                    ).contains('identity_ids', nostro.details).first()
                    if cor is None:
                        continue
                    accts = cor.get_accounts()
                    accts = [a for a in accts if a.ledger.identity_id == acct.details]
                    accts = [a for a in accts if a.type is AccountType.VOSTRO_LIABILITY]
                    if len(accts) < 1:
                        return False
                    vostro = accts[0]
                    if vostro.id not in [a.id for a in accounts]:
                        # each nostro Entry must have an offsetting vostro Entry
                        return False
                    offsetting_entry = [e for e in self.entries if e.account_id == vostro.id]
                    if len(offsetting_entry) < 1:
                        return False
                    if offsetting_entry[0].amount != entry.amount:
                        return False

                if acct.type is AccountType.VOSTRO_LIABILITY and type(acct.details) is str \
                    and acct.details != acct.id and Identity.find(acct.details):
                    # Vostro account must have equivalent Nostro account
                    vostro = acct
                    vostro.ledger().reload()
                    cor: Correspondence = Correspondence.query().contains(
                        'identity_ids', vostro.ledger.identity_id
                    ).contains('identity_ids', vostro.details).first()
                    if cor is None:
                        continue
                    accts = cor.get_accounts()
                    accts = [a for a in accts if a.ledger.identity_id == acct.details]
                    accts = [a for a in accts if a.type is AccountType.NOSTRO_ASSET]
                    if len(accts) < 1:
                        return False
                    nostro = accts[0]
                    if nostro.id not in [a.id for a in accounts]:
                        # each nostro Entry must have an offsetting nostro Entry
                        return False
                    offsetting_entry = [e for e in self.entries if e.account_id == nostro.id]
                    if len(offsetting_entry) < 1:
                        return False
                    if offsetting_entry[0].amount != entry.amount:
                        return False

        return True

    def save(self, tapescript_runtime: dict = {}, reload: bool = False) -> ArchivedTransaction:
        """Validate the transaction, save the entries, then save the
            transaction.
        """
        assert self.validate(tapescript_runtime, reload), 'cannot save an invalid Transaction'
        for e in self.entries:
            e.save()
        return super().save()



from __future__ import annotations
from sqloquent import HashedModel, RelatedModel, RelatedCollection, QueryBuilderProtocol
from .ArchivedEntry import ArchivedEntry
from .EntryType import EntryType
from typing import Callable
import packify


class Entry(HashedModel):
    connection_info: str = ''
    table: str = 'entries'
    id_column: str = 'id'
    columns: tuple[str] = ('id', 'type', 'amount', 'nonce', 'account_id', 'details')
    id: str
    type: str
    amount: int
    nonce: bytes
    account_id: str
    details: bytes
    account: RelatedModel
    transactions: RelatedCollection

    def __hash__(self) -> int:
        data = self.encode_value(self._encode(self.data))
        return hash(bytes(data, 'utf-8'))

    # override automatic properties
    @property
    def type(self) -> EntryType:
        """The EntryType of the Entry."""
        return EntryType(self.data['type'])
    @type.setter
    def type(self, val: EntryType):
        if type(val) is not EntryType:
            return
        self.data['type'] = val.value

    @property
    def details(self) -> packify.SerializableType:
        """A packify.SerializableType stored in the database as a blob."""
        return packify.unpack(self.data.get('details', b'M@\x00'))
    @details.setter
    def details(self, val: packify.SerializableType):
        self.data['details'] = packify.pack(val)

    @staticmethod
    def _encode(data: dict|None) -> dict|None:
        if type(data) is not dict:
            return data
        if type(data.get('type', None)) is EntryType:
            data['type'] = data['type'].value
        if type(data.get('details', {})) is not bytes:
            data['details'] = packify.pack(data.get('details', None))
        return data

    @classmethod
    def generate_id(cls, data: dict) -> str:
        """Generate an id by hashing the non-id contents. Raises
            TypeError for unencodable type (calls packify.pack).
        """
        return super().generate_id(cls._encode({**data}))

    @classmethod
    def insert(cls, data: dict) -> Entry | None:
        """Ensure data is encoded before inserting."""
        result = super().insert(cls._encode(data))
        return result

    @classmethod
    def insert_many(cls, items: list[dict]) -> int:
        """Ensure data is encoded before inserting."""
        items = [cls._encode(data) for data in list]
        return super().insert_many(items)

    @classmethod
    def query(cls, conditions: dict = None) -> QueryBuilderProtocol:
        """Ensure conditions are encoded properly before querying."""
        return super().query(cls._encode(conditions))

    @classmethod
    def set_sigfield_plugin(cls, plugin: Callable):
        """Sets the plugin function used by self.get_sigfields that
            parses the Entry to extract the correct sigfields for
            tapescript authorization. This is an optional override.
        """
        cls._plugin = plugin

    def get_sigfields(self, *args, **kwargs) -> dict[str, bytes]:
        """Get the sigfields for tapescript authorization. By default,
            it returns {sigfield1: self.generate_id()} because the ID
            cryptographically commits to all record data. If the
            set_sigfield_plugin method was previously called, this will
            instead return the result of calling the plugin function.
        """
        if hasattr(self, '_plugin') and callable(self._plugin):
            return self._plugin(self, *args, **kwargs)
        return {'sigfield1': bytes.fromhex(self.generate_id({**self.data}))}

    def archive(self) -> ArchivedEntry|None:
        """Archive the Entry. If it has already been archived,
            return the existing ArchivedEntry.
        """
        archived_entry_id = ArchivedEntry.generate_id({**self.data})
        try:
            return ArchivedEntry.insert({**self.data})
        except Exception as e:
            return ArchivedEntry.find(archived_entry_id)



from __future__ import annotations
from .Correspondence import Correspondence
from .Entry import Entry, EntryType, ArchivedEntry
from .Ledger import Ledger
from .Transaction import Transaction, ArchivedTransaction
from merkleasy import Tree
from sqloquent.asyncql import (
    AsyncDeletedModel,
    AsyncHashedModel,
    AsyncRelatedCollection,
    AsyncRelatedModel,
    AsyncSqlQueryBuilder,
)
from sqloquent.errors import tert, vert
from time import time
import packify
import tapescript


class TxRollup(AsyncHashedModel):
    """A Transaction Roll-up is a collection of Transactions that have
        been consolidated: the IDs of the committed Transactions are the
        leaves of a Merkle tree, and the aggregate effects of the
        Transactions are maintained in a dict mapping account IDs to
        tuples of EntryType and int balances. A TxRollup created for a
        Correspondence must have a valid auth_script that unlocks the
        txru_lock in the Correspondence's details (or an n-of-n multisig
        lock made from the pubkeys of the Correspondence's identities if
        no txru_lock was saved). The height of a TxRollup is the number
        of TxRollups in its chain -- they form a blockchain of TxRollups,
        hence the inclusion of a parent_id. Only one child TxRollup can
        be added for a given parent TxRollup, and the balances of the
        child TxRollup are the sum of the effects of the Transactions in
        the child TxRollup and the parent TxRollup balances; i.e. the
        most recent TxRollup is the sum of all the Transactions
        committed to in previous TxRollups in the chain. Inclusion of a
        Transaction can only be proven using the Merkle tree of the
        TxRollup in which it was committed and only if the full list of
        tx_ids is saved, but the proof can be verified by mirrors that
        have only the tx_root.
    """
    connection_info: str = ''
    table: str = 'txn_rollups'
    id_column: str = 'id'
    columns: tuple[str] = (
        'id', 'height', 'parent_id', 'tx_ids', 'tx_root', 'correspondence_id',
        'ledger_id', 'balances', 'timestamp', 'auth_script'
    )
    columns_excluded_from_hash: tuple[str] = ('tx_ids', 'auth_script')
    id: str
    height: int
    parent_id: str|None
    tx_ids: str
    tx_root: str
    correspondence_id: str|None
    ledger_id: str|None
    balances: bytes
    timestamp: str
    auth_script: bytes|None
    correspondence: AsyncRelatedModel
    ledger: AsyncRelatedModel
    transactions: AsyncRelatedCollection
    parent: AsyncRelatedModel
    child: AsyncRelatedModel

    def public(self) -> dict:
        """Returns the public data for mirroring this TxRollup. Excludes
            the tx_ids.
        """
        return {
            k:v for k,v in self.data.items()
            if k != 'tx_ids'
        }

    @property
    def tx_ids(self) -> list[str]:
        """A list of transaction IDs. Setting causes the ids to be
            sorted, then combined into a Merkle Tree, the root of which
            is used to set `self.tx_root`.
        """
        return self.data.get('tx_ids', '').split(',')
    @tx_ids.setter
    def tx_ids(self, val: list[str]):
        if type(val) is not list:
            return
        # sort the ids, join into a comma-separated string
        val.sort()
        self.data['tx_ids'] = ','.join(val)
        # convert to bytes and build a merkle tree
        val = [bytes.fromhex(txn_id) for txn_id in val]
        while len(val) < 2:
            val = [b'\x00'*32, *val]
        tree = Tree.from_leaves(val)
        self.data['tx_root'] = tree.root.hex()

    @property
    def balances(self) -> dict[str, tuple[EntryType, int]]:
        """A dict mapping account IDs to tuple[EntryType, int] balances."""
        balances: dict = packify.unpack(self.data.get('balances', b'M@\x00'))
        return {
            k: (EntryType(v[0]), v[1])
            for k, v in balances.items()
        }
    @balances.setter
    def balances(self, val: dict[str, tuple[EntryType, int]]):
        tert(type(val) is dict, 'balances must be a dict')
        tert(all([
                type(k) is str and
                type(v) is tuple and
                len(v) == 2 and
                type(v[0]) is EntryType and
                type(v[1]) is int
                for k, v in val.items()
            ]),
            'balances must be a dict of str: tuple[EntryType, int]')
        val = {
            k: (v[0].value, v[1])
            for k, v in val.items()
        }
        self.data['balances'] = packify.pack(val)

    @property
    def tree(self) -> Tree:
        """A merkle tree of the transaction IDs."""
        return Tree.from_leaves([bytes.fromhex(txn_id) for txn_id in self.tx_ids])

    def prove_txn_inclusion(self, txn_id: str|bytes) -> bytes:
        """Proves that a transaction is included in the tx rollup."""
        txn_id = bytes.fromhex(txn_id) if type(txn_id) is str else txn_id
        return self.tree.prove(txn_id)

    def verify_txn_inclusion_proof(self, txn_id: str|bytes, proof: bytes) -> bool:
        """Verifies that a transaction is included in the tx rollup."""
        txn_id = bytes.fromhex(txn_id) if type(txn_id) is str else txn_id
        return Tree.verify(bytes.fromhex(self.tx_root), txn_id, proof)

    @classmethod
    async def calculate_balances(
        cls, txns: list[Transaction],
        parent_balances: dict[str, tuple[EntryType, int]]|None = None,
        reload: bool = False
    ) -> dict[str, tuple[EntryType, int]]:
        """Calculates the account balances for a list of rolled-up
            transactions. If parent_balances is provided, those are the
            starting balances to which the balances of the rolled-up
            transactions are added. If reload is True, the entries are
            reloaded from the database.
        """
        balances = parent_balances or {}
        for txn in txns:
            if reload:
                await txn.entries().reload()
            for e in txn.entries:
                e: Entry
                bal = {EntryType.CREDIT: 0, EntryType.DEBIT: 0}
                if e.account_id in balances:
                    bal[balances[e.account_id][0]] = balances[e.account_id][1]
                bal[e.type] += e.amount
                net_credit = bal[EntryType.CREDIT] - bal[EntryType.DEBIT]
                if net_credit >= 0:
                    balances[e.account_id] = (EntryType.CREDIT, net_credit)
                else:
                    balances[e.account_id] = (EntryType.DEBIT, -net_credit)
        return balances

    @classmethod
    async def prepare(cls, txns: list[Transaction], parent_id: str|None = None,
                correspondence: Correspondence|None = None,
                ledger: Ledger|None = None, reload: bool = False
                ) -> TxRollup:
        """Prepare a tx rollup by checking that all txns are for the
            accounts of the given correspondence or belong to the same
            ledger if no correspondence is provided. Raises TypeError if
            txns is not a list of Transaction objects. Raises ValueError
            if any txns are not for accounts of the given correspondence
            or of the same ledger if no correspondence is provided, or
            if the parent TxRollup already has a child, or if there are
            no txns and no ledger or correspondence is provided, or if
            a TxRollup chain already exists for the given ledger or
            correspondence when no parent is provided. The Transaction
            IDs are sorted and combined into a Merkle Tree, the root of
            which is used to set the `tx_root` property.
        """
        tert(all([type(t) is Transaction for t in txns]),
            'txns must be a list of Transaction objects')
        tert(type(correspondence) is Correspondence or correspondence is None,
            'correspondence must be a Correspondence object or None')
        vert(len(txns) > 0 or ledger is not None or correspondence is not None,
            'either txns, ledger, or correspondence must be provided')

        accounts = []
        acct_ids = set()
        balances = {}
        txru = TxRollup()
        txru.tx_ids = [t.id for t in txns]
        txru.height = 0

        if correspondence is None:
            # all txns must be accounts from the same ledger
            if len(txns) > 0 and ledger is None:
                await txns[0].entries().reload()
                await txns[0].entries[0].account().reload()
                ledger: Ledger = txns[0].entries[0].account.ledger
            if reload:
                await ledger.accounts().reload()
            accounts = list(ledger.accounts)
            acct_ids = set([a.id for a in accounts])
            for txn in txns:
                await txn.entries().reload()
                for e in txn.entries:
                    e: Entry
                    vert(e.account_id in acct_ids,
                        'all txns must be for from the same ledger when correspondence is None')
        else:
            # all txns must be for accounts from the same correspondence
            accounts = await correspondence.get_accounts()
            accounts = [a for _, aa in accounts.items() for _, a in aa.items()]
            acct_ids = set([a.id for a in accounts])
            for txn in txns:
                await txn.entries().reload()
                for e in txn.entries:
                    e: Entry
                    vert(e.account_id in acct_ids,
                        'all txns must be for accounts from the same correspondence')

        if parent_id is not None:
            # if there is a parent, get its balances and set the height
            parent: TxRollup|None = await TxRollup.find(parent_id)
            vert(parent is not None, 'parent must exist')
            balances = parent.balances
            txru.height = parent.height + 1
            vert((await parent.child().query().count()) == 0, 'parent already has a child')
        else:
            # if there is no parent, ensure there is no other chain
            if ledger is not None:
                vert(await TxRollup.query().equal('ledger_id', ledger.id).count() == 0,
                    'the given ledger already has a TxRollup chain')
            elif correspondence is not None:
                vert(await TxRollup.query().equal('correspondence_id', correspondence.id).count() == 0,
                    'the given correspondence already has a TxRollup chain')

        # aggregate balances from txn entries
        balances = await cls.calculate_balances(txns, balances, reload=reload)

        txru.parent_id = parent_id
        txru.balances = balances
        txru.timestamp = str(time())
        if correspondence is not None:
            txru.correspondence_id = correspondence.id
        else:
            if len(txns) > 0 and ledger is None:
                ledger: Ledger = txns[0].entries[0].account.ledger
            txru.ledger_id = ledger.id
        return txru

    async def validate(self, reload: bool = False) -> bool:
        """Validates that a TxRollup has been authorized properly; that
            the balances are correct; and that the height is 1 + the
            height of the parent tx rollup (if one exists); and that
            there is no other chain for the relevant ledger or
            correspondence when no parent is provided.
        """
        authorized = True
        balances = {}
        parent = None

        # if there is a parent, get its balances if it is a tx rollup
        if self.parent_id is not None:
            parent: TxRollup|None = await TxRollup.find(self.parent_id)
            vert(parent is not None, 'parent must exist')
            balances = parent.balances
            await parent.child().reload()
            self_id = self.id or self.generate_id(self.data)
            if parent.child.id is not None:
                if parent.child.id != self_id:
                    return False

        if self.correspondence_id is not None:
            correspondence: Correspondence = await Correspondence.find(self.correspondence_id)
            await correspondence.identities().reload()
            # either the txru_lock has been set and fulfilled, or both
            # identities have signed independently
            txru_lock = correspondence.txru_lock
            if txru_lock is None:
                pubkeys = [identity.pubkey for identity in correspondence.identities]
                # if not all identities have a pubkey and the txru_lock
                # is not set, then the txru is authorized by default
                if not all([len(pk) > 0 for pk in pubkeys]):
                    authorized = True
                else:
                    txru_lock = tapescript.make_multisig_lock(pubkeys, len(pubkeys)).bytes

            if self.auth_script is not None and txru_lock is not None:
                authorized = tapescript.run_auth_scripts(
                    [self.auth_script, txru_lock],
                    {'sigfield1': self.id}
                )

        # validate the height
        if parent is None:
            if self.height != 0:
                return False
        else:
            if parent.height + 1 != self.height:
                return False

        # ensure there is no other chain
        if parent is None:
            self_id = self.id or self.generate_id(self.data)
            if self.correspondence_id is not None:
                if await TxRollup.query().equal(
                    'correspondence_id',
                    self.correspondence_id
                ).not_equal('id', self_id).count() > 0:
                    return False
            elif self.ledger_id is not None:
                if await TxRollup.query().equal(
                    'ledger_id',
                    self.ledger_id
                ).not_equal('id', self_id).count() > 0:
                    return False
            elif len(self.tx_ids) > 0:
                txn: Transaction|None = await Transaction.find(self.tx_ids[0])
                if txn is not None:
                    await txn.entries().reload()
                    if len(txn.entries) > 0:
                        await txn.entries[0].account().reload()
                        ledger_id = txn.entries[0].account.ledger_id
                        if await TxRollup.query().equal(
                            'ledger_id',
                            ledger_id
                        ).not_equal('id', self_id).count() > 0:
                            return False

        # recalculate the balances
        balances = await self.calculate_balances(self.transactions, balances, reload=reload)

        # compare the recalculated balances to the stored balances
        for acct_id, (entry_type, amount) in balances.items():
            if acct_id not in self.balances:
                return False
            if self.balances[acct_id][0] != entry_type:
                return False
            if self.balances[acct_id][1] != amount:
                return False

        return authorized

    async def trim(self, archive: bool = True) -> int:
        """Trims the transactions and entries committed to in this tx
            rollup. Returns the number of transactions trimmed. If
            archive is True, the transactions and entries are archived
            before being deleted. Raises ValueError if the tx rollup is
            not valid.
        """
        vert(await self.validate(), 'tx rollup is not valid')
        await self.transactions().reload()
        txns = self.transactions
        for txn in txns:
            txn: Transaction
            await txn.entries().reload()
            for e in txn.entries:
                e: Entry
                if archive:
                    await e.archive()
                await e.delete()
            if archive:
                await txn.archive()
            await txn.delete()
        return len(txns)

    def trimmed_transactions(self) -> AsyncSqlQueryBuilder:
        """Returns a query builder for AsyncDeletedModels containing the
            trimmed transactions committed to in this tx rollup.
        """
        return AsyncDeletedModel.query({'model_class': Transaction.__name__}).is_in(
            'record_id', self.tx_ids
        )

    async def trimmed_entries(self) -> AsyncSqlQueryBuilder:
        """Returns a query builder for AsyncDeletedModels containing the
            trimmed entries from trimmed transactions committed to in
            this tx rollup.
        """
        txns = [
            Transaction(packify.unpack(item.record))
            for item in await self.trimmed_transactions().get()
        ]
        return AsyncDeletedModel.query({'model_class': Entry.__name__}).is_in(
            'record_id',
            [
                eid
                for txn in txns
                for eid in txn.entry_ids.split(',')
            ]
        )

    def archived_transactions(self) -> AsyncSqlQueryBuilder:
        """Returns a query builder for ArchivedTransactions committed
            to in this tx rollup.
        """
        return ArchivedTransaction.query().is_in('id', self.tx_ids)

    async def archived_entries(self) -> AsyncSqlQueryBuilder:
        """Returns a query builder for ArchivedEntries committed to
            in this tx rollup.
        """
        return ArchivedEntry.query().is_in(
            'id',
            [
                e.id
                for txn in await self.archived_transactions().get()
                for e in txn.entries
            ]
        )



from .models import (
    Account,
    AccountCategory,
    AccountType,
    ArchivedEntry,
    ArchivedTransaction,
    Correspondence,
    Currency,
    Customer,
    DeletedModel,
    Entry,
    EntryType,
    Identity,
    Ledger,
    LedgerType,
    Transaction,
    TxRollup,
    Vendor,
    set_connection_info,
    get_migrations,
    publish_migrations,
    automigrate,
)
from .version import version




from context import models
from genericpath import isfile
from nacl.signing import SigningKey
from packify import pack, unpack
from time import time
import os
import sqloquent.tools
import tapescript
import unittest


DB_FILEPATH = 'tests/test.db'
MIGRATIONS_PATH = 'tests/migrations'
MODELS_PATH = 'bookchain/models'


class TestTxRollupE2E(unittest.TestCase):
    @classmethod
    def automigrate(cls):
        sqloquent.tools.publish_migrations(MIGRATIONS_PATH)
        tomigrate = [
            models.Identity, models.Currency, models.Ledger,
            models.Account, models.Entry, models.Transaction,
            models.Correspondence, models.TxRollup,
            models.ArchivedTransaction, models.ArchivedEntry
        ]
        for model in tomigrate:
            name = model.__name__
            m = sqloquent.tools.make_migration_from_model(model, name)
            with open(f'{MIGRATIONS_PATH}/create_{name}.py', 'w') as f:
                f.write(m)
        sqloquent.tools.automigrate(MIGRATIONS_PATH, DB_FILEPATH)

    @classmethod
    def setUpClass(cls):
        models.Identity.connection_info = DB_FILEPATH
        models.Currency.connection_info = DB_FILEPATH
        models.Correspondence.connection_info = DB_FILEPATH
        models.Ledger.connection_info = DB_FILEPATH
        models.Account.connection_info = DB_FILEPATH
        models.Entry.connection_info = DB_FILEPATH
        models.Transaction.connection_info = DB_FILEPATH
        models.TxRollup.connection_info = DB_FILEPATH
        models.ArchivedTransaction.connection_info = DB_FILEPATH
        models.ArchivedEntry.connection_info = DB_FILEPATH
        sqloquent.DeletedModel.connection_info = DB_FILEPATH
        cls.automigrate()
        super().setUpClass()

    def setUp(self):
        models.Identity.query().delete()
        models.Currency.query().delete()
        models.Correspondence.query().delete()
        models.Ledger.query().delete()
        models.Account.query().delete()
        models.Entry.query().delete()
        models.Transaction.query().delete()
        models.TxRollup.query().delete()
        models.ArchivedTransaction.query().delete()
        models.ArchivedEntry.query().delete()
        sqloquent.DeletedModel.query().delete()
        self.setup_cryptographic_values()
        super().setUp()

    def tearDown(self):
        for file in os.listdir(MIGRATIONS_PATH):
            if isfile(f'{MIGRATIONS_PATH}/{file}'):
                os.remove(f'{MIGRATIONS_PATH}/{file}')
        super().tearDown()

    @classmethod
    def tearDownClass(cls):
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        return super().tearDownClass()

    def setup_cryptographic_values(self):
        self.seed_alice = os.urandom(32)
        self.seed_bob = os.urandom(32)
        self.pkey_alice = bytes(SigningKey(self.seed_alice).verify_key)
        self.pkey_bob = bytes(SigningKey(self.seed_bob).verify_key)
        self.committed_script_alice = tapescript.tools.make_delegate_key_lock(self.pkey_alice)
        self.committed_script_bob = tapescript.tools.make_delegate_key_lock(self.pkey_bob)
        self.locking_script_alice = tapescript.tools.make_taproot_lock(
            self.pkey_alice,
            self.committed_script_alice
        ).bytes
        self.locking_script_bob = tapescript.tools.make_taproot_lock(
            self.pkey_bob,
            self.committed_script_bob
        ).bytes
        self.delegate_seed = os.urandom(32) # for simplicity, both will have the same delegate
        self.delegate_pkey = bytes(SigningKey(self.delegate_seed).verify_key)
        begin_ts = int(time()) - 1
        end_ts = int(time()) + 60*60*24*365
        self.delegate_cert_alice = tapescript.make_delegate_key_cert(
            self.seed_alice, self.delegate_pkey, begin_ts, end_ts
        )
        self.delegate_cert_bob = tapescript.make_delegate_key_cert(
            self.seed_bob, self.delegate_pkey, begin_ts, end_ts
        )
        self.multisig_lock = tapescript.tools.make_multisig_lock(
            [self.pkey_alice, self.pkey_bob], 2
        ).bytes

    def setup_currency(self) -> models.Currency:
        # set up currency
        self.currency = models.Currency.insert({
            'name': 'Median Human Hour',
            'prefix_symbol': 'Ħ',
            'fx_symbol': 'MHH',
            'base': 60,
            'unit_divisions': 2,
            'details': 'Abstract value of one median hour of human time. ' +
                '1 Hour = 60 Minutes = 3600 Seconds',
        })
        return self.currency

    def setup_identities(self) -> tuple[models.Identity, models.Identity]:
        # set up alice, ledger_alice, accounts, and starting capital transaction
        alice: models.Identity = models.Identity.insert({
            'name': 'Alice',
            'pubkey': self.pkey_alice,
            'seed': self.seed_alice,
        })
        ledger_alice = models.Ledger.insert({
            'name': 'Current Ledger',
            'identity_id': alice.id,
            'currency_id': self.currency.id,
        })
        asset_acct_alice, liability_acct_alice, equity_acct_alice = ledger_alice.setup_basic_accounts()

        equity_acct_alice.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_alice
        }
        equity_acct_alice.save()

        asset_acct_alice.locking_scripts = {
            models.EntryType.CREDIT: self.locking_script_alice
        }
        asset_acct_alice.save()

        liability_acct_alice.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_alice,
            models.EntryType.CREDIT: self.locking_script_alice,
        }
        liability_acct_alice.save()

        # fund the Identity with starting capital
        nonce = os.urandom(16)
        equity_entry = models.Entry({
            'type': models.EntryType.CREDIT,
            'account_id': equity_acct_alice.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        equity_entry.account = equity_acct_alice
        asset_entry = models.Entry({
            'type': models.EntryType.DEBIT,
            'account_id': asset_acct_alice.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        asset_entry.account = asset_acct_alice
        txn = models.Transaction.prepare([equity_entry, asset_entry], str(time()))
        equity_entry.save()
        asset_entry.save()
        txn.save()

        # set up bob, ledger_bob, and some accounts
        bob: models.Identity = models.Identity.insert({
            'name': 'Bob',
            'pubkey': self.pkey_bob,
            'seed': self.seed_bob,
        })
        ledger_bob = models.Ledger.insert({
            'name': 'Current Ledger',
            'identity_id': bob.id,
            'currency_id': self.currency.id,
        })
        asset_acct_bob, liability_acct_bob, equity_acct_bob = ledger_bob.setup_basic_accounts()
        equity_acct_bob.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_bob
        }
        equity_acct_bob.save()
        asset_acct_bob.locking_scripts = {
            models.EntryType.CREDIT: self.locking_script_bob
        }
        asset_acct_bob.save()
        liability_acct_bob.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_bob,
            models.EntryType.CREDIT: self.locking_script_bob,
        }
        liability_acct_bob.save()

        # fund the Identity with starting capital
        nonce = os.urandom(16)
        equity_entry = models.Entry({
            'type': models.EntryType.CREDIT,
            'account_id': equity_acct_bob.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        equity_entry.account = equity_acct_bob
        asset_entry = models.Entry({
            'type': models.EntryType.DEBIT,
            'account_id': asset_acct_bob.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        asset_entry.account = asset_acct_bob
        txn = models.Transaction.prepare([equity_entry, asset_entry], str(time()))
        equity_entry.save()
        asset_entry.save()
        txn.save()

        return alice, bob

    def setup_correspondence(self) -> models.Correspondence:
        alice: models.Identity = models.Identity.query().equal(name='Alice').first()
        bob: models.Identity = models.Identity.query().equal(name='Bob').first()
        ledger_alice: models.Ledger = models.Ledger.query().equal(identity_id=alice.id).first()
        ledger_bob: models.Ledger = models.Ledger.query().equal(identity_id=bob.id).first()

        # set up Correspondence
        assert alice.correspondences().query().count() == 0
        assert len(alice.correspondents(reload=True)) == 0
        correspondence: models.Correspondence = models.Correspondence.insert({
            'identity_ids': ','.join(sorted([alice.id, bob.id])),
            'ledger_ids': ','.join(sorted([ledger_alice.id, ledger_bob.id])),
            'details': pack({
                'starting_capital': 'Ħ2000',
                'limits': {
                    alice.id: 'Ħ166 M40',
                    bob.id: 'Ħ166 M40',
                },
                'locking_scripts': {
                    alice.id: self.locking_script_alice,
                    bob.id: self.locking_script_bob,
                },
            })
        })
        assert alice.correspondences().query().count() == 1
        assert len(alice.correspondents(reload=True)) == 1
        assert alice.correspondents()[0].id == bob.id
        assert bob.correspondences().query().count() == 1
        assert len(bob.correspondents(reload=True)) == 1
        assert bob.correspondents()[0].id == alice.id

        # set up correspondence accounts
        cor_accts = correspondence.setup_accounts({
            alice.id: self.locking_script_alice,
            bob.id: self.locking_script_bob,
        })
        for _, acct in cor_accts[alice.id].items():
            acct.save()
        for _, acct in cor_accts[bob.id].items():
            acct.save()

        return correspondence

    def create_txn(self, acct1: models.Account, acct2: models.Account, amount: int) -> models.Transaction:
        nonce = os.urandom(16)
        entries = [
            models.Entry({
                'type': models.EntryType.DEBIT,
                'account_id': acct1.id,
                'amount': amount,
                'nonce': nonce,
            }),
            models.Entry({
                'type': models.EntryType.CREDIT,
                'account_id': acct2.id,
                'amount': amount,
                'nonce': nonce,
            }),
        ]
        txn = models.Transaction.prepare(entries, str(time()))
        txn.save()
        return txn

    def test_simple_e2e(self):
        self.setup_currency()
        alice, _ = self.setup_identities()
        ledger: models.Ledger = alice.ledgers[0]

        asset_acct: models.Account = [acct for acct in ledger.accounts if acct.type == models.AccountType.ASSET][0]
        equity_acct: models.Account = [acct for acct in ledger.accounts if acct.type == models.AccountType.EQUITY][0]
        asset_starting_balance = asset_acct.balance()
        equity_starting_balance = equity_acct.balance()

        # create txns: Alice adds assets to her ledger
        txn1 = self.create_txn(asset_acct, equity_acct, 100)
        txn2 = self.create_txn(asset_acct, equity_acct, 200)

        # ensure the balances are correct
        assert asset_acct.balance() == asset_starting_balance + 300, \
            f'{asset_acct.balance()} != {asset_starting_balance} + 300'
        assert equity_acct.balance() == equity_starting_balance + 300, \
            f'{equity_acct.balance()} != {equity_starting_balance} + 300'

        # test empty txrollup
        (models.TxRollup()).balances
        (models.TxRollup()).tx_ids

        # create a txrollup
        txrollup = models.TxRollup.prepare([txn1, txn2])
        assert txrollup.validate()
        txrollup.save()

        # prove inclusion of txn
        proof = txrollup.prove_txn_inclusion(txn1.id)
        assert txrollup.verify_txn_inclusion_proof(txn1.id, proof)
        proof = txrollup.prove_txn_inclusion(txn2.id)
        assert txrollup.verify_txn_inclusion_proof(txn2.id, proof)

        # test empty ArchivedTransaction and ArchivedEntry
        (models.ArchivedTransaction()).details
        (models.ArchivedTransaction()).auth_scripts
        (models.ArchivedEntry()).details

        # archive and trim txn and entries
        assert txrollup.trim() == 2
        assert txrollup.archived_transactions().count() == 2, txrollup.archived_transactions().count()
        assert txrollup.archived_entries().count() == 4, txrollup.archived_entries().count()

        # ensure the txn has been trimmed
        assert txrollup.trimmed_transactions().count() == 2
        assert txrollup.trimmed_entries().count() == 4
        assert models.Transaction.query().is_in('id', txrollup.tx_ids).count() == 0
        entry_ids = [e.id for e in txn1.entries] + [e.id for e in txn2.entries]
        assert models.Entry.query().is_in('id', entry_ids).count() == 0

        # ensure the balances are correct
        assert asset_acct.balance() == asset_starting_balance # without rolled up balances
        assert asset_acct.balance(rolled_up_balances=txrollup.balances) == asset_starting_balance + 300, \
            f'{asset_acct.balance(rolled_up_balances=txrollup.balances)} != {asset_starting_balance} + 300'
        assert equity_acct.balance() == equity_starting_balance # without rolled up balances
        assert equity_acct.balance(rolled_up_balances=txrollup.balances) == equity_starting_balance + 300, \
            f'{equity_acct.balance(rolled_up_balances=txrollup.balances)} != {equity_starting_balance} + 300'

        # mirror the txrollup's public data
        public_data = txrollup.public()
        mirrored = models.TxRollup(public_data)
        assert mirrored.validate()
        # inclusion proofs should be verified by mirrored TxRollup
        proof = txrollup.prove_txn_inclusion(txn2.id)
        assert mirrored.verify_txn_inclusion_proof(txn2.id, proof)

        # create more txns and txrollups
        txn3 = self.create_txn(asset_acct, equity_acct, 10)
        txn4 = self.create_txn(asset_acct, equity_acct, 20)
        txrollup2 = models.TxRollup.prepare([txn3, txn4], txrollup.id)
        assert txrollup2.validate()
        txrollup2.save()

        # attempt to create a competing txrollup chain
        with self.assertRaises(ValueError) as e:
            models.TxRollup.prepare([txn3, txn4])
        assert str(e.exception) == 'the given ledger already has a TxRollup chain'

        # prove inclusion of txn
        proof = txrollup2.prove_txn_inclusion(txn3.id)
        assert txrollup2.verify_txn_inclusion_proof(txn3.id, proof)
        proof = txrollup2.prove_txn_inclusion(txn4.id)
        assert txrollup2.verify_txn_inclusion_proof(txn4.id, proof)

        # archive and trim txn and entries
        assert txrollup2.trim() == 2
        assert txrollup2.archived_transactions().count() == 2, txrollup2.archived_transactions().count()
        assert txrollup2.archived_entries().count() == 4, txrollup2.archived_entries().count()

        # ensure the balances are correct
        assert asset_acct.balance(rolled_up_balances=txrollup2.balances) == asset_starting_balance + 330, \
            f'{asset_acct.balance(rolled_up_balances=txrollup2.balances)} != {asset_starting_balance} + 330'
        assert equity_acct.balance(rolled_up_balances=txrollup2.balances) == equity_starting_balance + 330, \
            f'{equity_acct.balance(rolled_up_balances=txrollup2.balances)} != {equity_starting_balance} + 330'

        # check relations
        assert len(ledger.rollups) == 2
        assert txrollup.id in [r.id for r in ledger.rollups]
        assert txrollup2.id in [r.id for r in ledger.rollups]
        assert txrollup.ledger.id == ledger.id
        assert txrollup2.ledger.id == ledger.id
        assert len(ledger.archived_transactions) == 4
        archived_txn: models.ArchivedTransaction = ledger.archived_transactions[0]
        assert len(archived_txn.entries) == 2
        archived_entry: models.ArchivedEntry = archived_txn.entries[0]
        assert len(archived_entry.transactions) == 1
        assert archived_entry.transactions[0].id == archived_txn.id
        assert len(asset_acct.archived_entries) == 4

        # make some more txns
        txn5 = self.create_txn(asset_acct, equity_acct, 100)
        txn6 = self.create_txn(asset_acct, equity_acct, 200)

        # try to add a second child txrollup to the same parent
        with self.assertRaises(ValueError) as e:
            models.TxRollup.prepare([txn5, txn6], txrollup.id)
        assert str(e.exception) == 'parent already has a child'
        # fake it
        txrollup3 = models.TxRollup.prepare([txn5, txn6], txrollup2.id)
        txrollup3.parent_id = txrollup.id
        txrollup3.height = txrollup.height + 1
        txrollup3.balances = models.TxRollup.calculate_balances([txn5, txn6], txrollup.balances)
        assert not txrollup3.validate()
        # reset the parent_id and height to attempt to make a competing chain
        txrollup3.parent_id = None
        txrollup3.height = 0
        txrollup3.balances = models.TxRollup.calculate_balances([txn5, txn6])
        assert not txrollup3.validate()

        # prepare a txrollup with just one txn
        txrollup3 = models.TxRollup.prepare([txn5], txrollup2.id)
        assert txrollup3.validate()

        # prepare an empty txrollup
        txrollup3 = models.TxRollup.prepare([], txrollup2.id, ledger=ledger)
        assert txrollup3.validate()

    def test_with_correspondence_e2e(self):
        self.setup_currency()
        alice, bob = self.setup_identities()
        correspondence = self.setup_correspondence()

        cor_accts = correspondence.get_accounts()
        nostro_acct_alice = cor_accts[alice.id][models.AccountType.NOSTRO_ASSET]
        vostro_acct_alice = cor_accts[alice.id][models.AccountType.VOSTRO_LIABILITY]
        equity_acct_alice = cor_accts[alice.id][models.AccountType.EQUITY]
        nostro_acct_bob = cor_accts[bob.id][models.AccountType.NOSTRO_ASSET]
        vostro_acct_bob = cor_accts[bob.id][models.AccountType.VOSTRO_LIABILITY]
        equity_acct_bob = cor_accts[bob.id][models.AccountType.EQUITY]

        # create txn: Alice credits her ledger's vostro to pay Bob
        _, entries = correspondence.pay_correspondent(alice, bob, 200, os.urandom(16))
        txn = models.Transaction.prepare(entries, str(time()), auth_scripts={
            equity_acct_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_alice, entries[0].get_sigfields(), self.committed_script_alice
            ).bytes,
            vostro_acct_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_alice, entries[2].get_sigfields(), self.committed_script_alice
            ).bytes,
            nostro_acct_bob.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_alice, entries[3].get_sigfields(), self.committed_script_alice
            ).bytes,
        })
        txn.save()
        assert nostro_acct_bob.balance() == 200
        assert vostro_acct_alice.balance() == 200

        # create txn: Bob credits his ledger's nostro to pay Alice
        entries, _ = correspondence.pay_correspondent(bob, alice, 100, os.urandom(16))
        txn2 = models.Transaction.prepare(entries, str(time()), auth_scripts={
            equity_acct_bob.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_bob, entries[0].get_sigfields(), self.committed_script_bob
            ).bytes,
            nostro_acct_bob.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_bob, entries[2].get_sigfields(), self.committed_script_bob
            ).bytes,
            vostro_acct_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_bob, entries[3].get_sigfields(), self.committed_script_bob
            ).bytes,
        })
        txn2.save()
        assert nostro_acct_bob.balance() == 100
        assert vostro_acct_alice.balance() == 100

        # create a txrollup
        txrollup = models.TxRollup.prepare([txn, txn2], correspondence=correspondence)
        assert txrollup.validate()
        txrollup.save()

        # prove inclusion of each txn
        proof = txrollup.prove_txn_inclusion(txn.id)
        assert txrollup.verify_txn_inclusion_proof(txn.id, proof)
        proof = txrollup.prove_txn_inclusion(txn2.id)
        assert txrollup.verify_txn_inclusion_proof(txn2.id, proof)

        # archive and trim txns and entries
        assert txrollup.trim() == 2
        assert txrollup.archived_transactions().count() == 2, txrollup.archived_transactions().count()
        assert txrollup.archived_entries().count() == 8, txrollup.archived_entries().count()

        # ensure the txns have been trimmed
        assert txrollup.trimmed_transactions().count() == 2
        assert txrollup.trimmed_entries().count() == 8
        assert models.Transaction.query().is_in('id', txrollup.tx_ids).count() == 0
        entry_ids = [e.id for e in txn.entries] + [e.id for e in txn2.entries]
        assert models.Entry.query().is_in('id', entry_ids).count() == 0

        # prove inclusion of each trimmed txn
        txns = [models.Transaction(unpack(item.record)) for item in txrollup.trimmed_transactions().get()]
        for txn in txns:
            assert len(txn.id) > 0
            proof = txrollup.prove_txn_inclusion(txn.id)
            assert txrollup.verify_txn_inclusion_proof(txn.id, proof)

        # prove inclusion of each archived txn
        txns = txrollup.archived_transactions().get()
        for txn in txns:
            assert len(txn.id) > 0
            proof = txrollup.prove_txn_inclusion(txn.id)
            assert txrollup.verify_txn_inclusion_proof(txn.id, proof)

        # prove that the balances are correct when using rolled up balances
        assert nostro_acct_bob.balance() == 0
        assert nostro_acct_bob.balance(rolled_up_balances=txrollup.balances) == 100, \
            nostro_acct_bob.balance(rolled_up_balances=txrollup.balances)
        assert vostro_acct_alice.balance() == 0
        assert vostro_acct_alice.balance(rolled_up_balances=txrollup.balances) == 100, \
            vostro_acct_alice.balance(rolled_up_balances=txrollup.balances)
        balances = correspondence.balances()
        assert balances[alice.id] == 0, balances[alice.id]
        assert balances[bob.id] == 0, balances[bob.id]
        balances = correspondence.balances(rolled_up_balances=txrollup.balances)
        assert balances[alice.id] == -100, balances[alice.id]
        assert balances[bob.id] == 100, balances[bob.id]

        # check relations
        assert len(correspondence.rollups) == 1
        assert txrollup.id in [r.id for r in correspondence.rollups]
        assert txrollup.correspondence.id == correspondence.id


if __name__ == '__main__':
    unittest.main()



from context import models
from genericpath import isfile
from sqlite3 import OperationalError
from sqloquent import DeletedModel
from time import time
import os
import sqloquent.tools
import unittest


DB_FILEPATH = 'tests/test.db'
MIGRATIONS_PATH = 'tests/migrations'


class TestBasicE2E(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        models.Identity.connection_info = DB_FILEPATH
        models.Currency.connection_info = DB_FILEPATH
        models.Ledger.connection_info = DB_FILEPATH
        models.Account.connection_info = DB_FILEPATH
        models.AccountCategory.connection_info = DB_FILEPATH
        models.Entry.connection_info = DB_FILEPATH
        models.Transaction.connection_info = DB_FILEPATH
        sqloquent.DeletedModel.connection_info = DB_FILEPATH
        super().setUpClass()

    def setUp(self):
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        super().setUp()

    def tearDown(self):
        for file in os.listdir(MIGRATIONS_PATH):
            if isfile(f'{MIGRATIONS_PATH}/{file}'):
                os.remove(f'{MIGRATIONS_PATH}/{file}')
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        super().tearDown()

    def automigrate(self):
        sqloquent.tools.publish_migrations(MIGRATIONS_PATH)
        tomigrate = [
            models.Identity, models.Currency, models.Ledger,
            models.Account, models.AccountCategory, models.Entry,
            models.Transaction,
        ]
        for model in tomigrate:
            name = model.__name__
            m = sqloquent.tools.make_migration_from_model(model, name)
            with open(f'{MIGRATIONS_PATH}/create_{name}.py', 'w') as f:
                f.write(m)
        sqloquent.tools.automigrate(MIGRATIONS_PATH, DB_FILEPATH)

    def test_e2e(self):
        with self.assertRaises(OperationalError):
            models.Account.query().count()
        self.automigrate()
        assert models.Account.query().count() == 0

        # test empty Transaction
        (models.Transaction()).details
        (models.Transaction()).auth_scripts
        assert not models.Transaction().validate()

        # setup account categories
        equity_acct_cat = models.AccountCategory.insert({
            'name': 'Equity',
            'ledger_type': models.LedgerType.PRESENT,
            'destination': 'Balance Sheet',
        })
        assert equity_acct_cat is not None
        assert models.AccountCategory.find(equity_acct_cat.id) is not None
        asset_acct_cat = models.AccountCategory.insert({
            'name': 'Asset',
            'ledger_type': models.LedgerType.PRESENT,
            'destination': 'Balance Sheet',
        })
        liability_acct_cat = models.AccountCategory.insert({
            'name': 'Liability',
            'ledger_type': models.LedgerType.PRESENT,
            'destination': 'Balance Sheet',
        })

        # setup identity, currency, ledger, and some accounts
        identity = models.Identity.insert({'name': 'Test Man'})
        currency = models.Currency.insert({
            'name': 'US Dollar',
            'prefix_symbol': '$',
            'fx_symbol': 'USD',
            'base': 10,
            'unit_divisions': 2,
        })
        ledger = models.Ledger.insert({
            'name': 'General Ledger',
            'identity_id': identity.id,
            'currency_id': currency.id,
            'type': models.LedgerType.PRESENT,
        })
        equity_acct = models.Account.insert({
            'name': 'General Equity',
            'type': models.AccountType.EQUITY,
            'ledger_id': ledger.id,
            'category_id': equity_acct_cat.id,
        })
        asset_acct = models.Account.insert({
            'name': 'General Asset',
            'type': models.AccountType.ASSET,
            'ledger_id': ledger.id,
            'category_id': asset_acct_cat.id,
        })
        liability_acct = models.Account.insert({
            'name': 'General Liability',
            'type': models.AccountType.LIABILITY,
            'ledger_id': ledger.id,
            'category_id': liability_acct_cat.id,
        })

        assert equity_acct.category.id == equity_acct_cat.id
        assert asset_acct.category.id == asset_acct_cat.id
        assert liability_acct.category.id == liability_acct_cat.id

        assert len(liability_acct_cat.accounts) == 1, liability_acct_cat.accounts

        # make sub account
        assert len(liability_acct.children) == 0
        liability_sub_acct = models.Account.insert({
            'name': 'Liability Sub Account',
            'type': models.AccountType.LIABILITY,
            'ledger_id': ledger.id,
            'parent_id': liability_acct.id,
            'category_id': liability_acct_cat.id,
        })
        assert liability_sub_acct.parent is not None, liability_sub_acct.parent
        assert liability_sub_acct.parent.id == liability_acct.id, liability_sub_acct.parent
        liability_acct.children().reload()
        assert len(liability_acct.children) == 1
        assert liability_acct.children[0].id == liability_sub_acct.id

        # test empty Entry
        (models.Entry()).details

        # prepare and save a valid transaction
        txn_nonce = os.urandom(16)
        equity_entry = models.Entry({
            'type': models.EntryType.CREDIT,
            'account_id': equity_acct.id,
            'amount': 10_000_00,
            'nonce': txn_nonce,
        })
        equity_entry.account = equity_acct
        asset_entry = models.Entry({
            'type': models.EntryType.DEBIT,
            'account_id': asset_acct.id,
            'amount': 10_000_00,
            'nonce': txn_nonce,
        })
        asset_entry.account = asset_acct
        txn = models.Transaction.prepare(
            [equity_entry, asset_entry], str(time()),
            details='Starting capital asset'
        )
        assert txn.validate()
        txn.save()
        # reload txn from database and validate it
        txn: models.Transaction = models.Transaction.find(txn.id)
        assert txn.validate(reload=True)
        assert len(asset_entry.transactions) == 1
        assert asset_entry.transactions[0].id == txn.id, asset_entry.transactions

        # check balances
        assert equity_acct.balance() == 10_000_00, equity_acct.balance()
        assert asset_acct.balance() == 10_000_00, asset_acct.balance()
        assert liability_acct.balance() == 0, liability_acct.balance()
        balances = ledger.balances(True)
        assert len(balances.keys()) == 4, balances
        assert balances[equity_acct.id][0] == 10_000_00, balances[equity_acct.id][0]
        assert balances[asset_acct.id][0] == 10_000_00, balances[asset_acct.id][0]
        assert balances[liability_acct.id][0] == 0, balances[liability_acct.id][0]
        assert balances[liability_sub_acct.id][0] == 0, balances[liability_sub_acct.id][0]

        # prepare and save valid transaction for liability sub account
        txn_nonce = os.urandom(16)
        equity_entry = models.Entry.insert({
            'type': models.EntryType.DEBIT,
            'account_id': equity_acct.id,
            'amount': 9_99,
            'nonce': txn_nonce,
        })
        liability_entry = models.Entry.insert({
            'type': models.EntryType.CREDIT,
            'account_id': liability_sub_acct.id,
            'amount': 9_99,
            'nonce': txn_nonce,
        })
        txn = models.Transaction.prepare([equity_entry, liability_entry], str(time()))
        assert txn.validate()
        txn.save()

        # check balances
        assert equity_acct.balance() == 10_000_00-9_99, equity_acct.balance()
        assert asset_acct.balance() == 10_000_00, asset_acct.balance()
        assert liability_sub_acct.balance() == 9_99, liability_acct.balance()
        assert liability_acct.balance() == 9_99, liability_acct.balance()
        assert liability_acct.balance(False) == 0, liability_acct.balance(False)
        balances = ledger.balances(True)
        assert len(balances.keys()) == 4, balances
        assert balances[equity_acct.id][0] == 10_000_00-9_99, balances[equity_acct.id][0]
        assert balances[asset_acct.id][0] == 10_000_00, balances[asset_acct.id][0]
        assert balances[liability_acct.id][0] == 0, balances[liability_acct.id][0]
        assert balances[liability_sub_acct.id][0] == 9_99, balances[liability_sub_acct.id][0]

        # prepare invalid transaction: reused entries
        with self.assertRaises(ValueError) as e:
            txn = models.Transaction.prepare([equity_entry, asset_entry], str(int(time())))
        assert 'already contained within a Transaction' in str(e.exception)

        # prepare invalid transaction: unbalanced entries
        txn_nonce = os.urandom(16)
        equity_entry = models.Entry({
            'type': models.EntryType.CREDIT,
            'account_id': equity_acct.id,
            'amount': 10_00,
            'nonce': txn_nonce,
        })
        equity_entry.account = equity_acct
        asset_entry = models.Entry({
            'type': models.EntryType.DEBIT,
            'account_id': asset_acct.id,
            'amount': 10_01,
            'nonce': txn_nonce,
        })
        asset_entry.account = asset_acct
        with self.assertRaises(ValueError) as e:
            txn = models.Transaction.prepare([equity_entry, asset_entry], str(int(time())))
        assert 'unbalanced' in str(e.exception)

        # delete something
        deleted = identity.delete()
        assert isinstance(deleted, DeletedModel)
        assert models.Identity.find(identity.id) is None

        # restore deleted identity
        restored = deleted.restore({'Identity': models.Identity})
        assert isinstance(restored, models.Identity)
        assert restored.id == identity.id
        restored.save()
        assert models.Identity.find(identity.id) is not None


if __name__ == '__main__':
    unittest.main()



from context import models, bookchain, asyncql
from decimal import Decimal
from genericpath import isfile
from sqlite3 import OperationalError
from time import time
import os
import unittest


DB_FILEPATH = 'tests/test.db'
MIGRATIONS_PATH = 'tests/migrations'
MODELS_PATH = 'bookchain/models'


class TestMisc(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        bookchain.set_connection_info(DB_FILEPATH)
        super().setUpClass()

    def setUp(self):
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        super().setUp()

    def tearDown(self):
        for file in os.listdir(MIGRATIONS_PATH):
            if isfile(f'{MIGRATIONS_PATH}/{file}'):
                os.remove(f'{MIGRATIONS_PATH}/{file}')
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        super().tearDown()

    def test_set_connection_info(self):
        bookchain.set_connection_info(DB_FILEPATH)
        for name in dir(models):
            model = getattr(models, name)
            if hasattr(model, 'connection_info'):
                assert model.connection_info == DB_FILEPATH, model
        bookchain.set_connection_info('foobar')
        for name in dir(models):
            model = getattr(models, name)
            if hasattr(model, 'connection_info'):
                assert model.connection_info == 'foobar', model

        asyncql.set_connection_info(DB_FILEPATH)
        for name in dir(asyncql):
            model = getattr(asyncql, name)
            if hasattr(model, 'connection_info'):
                assert model.connection_info == DB_FILEPATH, model
        asyncql.set_connection_info('foobar')
        for name in dir(asyncql):
            model = getattr(asyncql, name)
            if hasattr(model, 'connection_info'):
                assert model.connection_info == 'foobar', model

    def test_currency(self):
        currency = models.Currency({
            'name': 'US Dollar',
            'prefix_symbol': '$',
            'fx_symbol': 'USD',
            'base': 100,
            'unit_divisions': 1,
        })

        amount = Decimal('1.23')
        assert currency.format(123) == '$1.23', currency.format(123)
        assert currency.get_units(123) == (1, 23), currency.get_units(123)
        assert currency.to_decimal(123) == amount, currency.to_decimal(123)
        assert currency.from_decimal(amount) == 123, currency.from_decimal(amount)

        currency = models.Currency({
            'name': 'Mean Minute/Hour',
            'prefix_symbol': 'Ħ',
            'fx_symbol': 'MMH',
            'base': 60,
            'unit_divisions': 2,
        })

        assert currency.format(60*60*1.23) == 'Ħ1.23', currency.format(60*60*1.23)
        assert currency.format(60*60 + 45, decimal_places=2) == 'Ħ1.01', \
            currency.format(60*60 + 45, decimal_places=2)
        assert currency.to_decimal(5011) == Decimal(5011)/Decimal(60*60)

        amount = (60**2)*2 + (60**1)*2 + (60**0)*3
        assert currency.get_units(amount) == (2, 2, 3)
        assert currency.format(amount, use_decimal=False, divider=':') == 'Ħ02:02:03', \
            currency.format(amount, use_decimal=False, divider=':')

        amount = Decimal('1.51')
        assert currency.from_decimal(amount) == 5436, currency.from_decimal(amount)

    def test_asyncql_currency(self):
        currency = asyncql.Currency({
            'name': 'US Dollar',
            'prefix_symbol': '$',
            'fx_symbol': 'USD',
            'base': 100,
            'unit_divisions': 1,
        })

        amount = Decimal('1.23')
        assert currency.format(123) == '$1.23', currency.format(123)
        assert currency.get_units(123) == (1, 23), currency.get_units(123)
        assert currency.to_decimal(123) == amount, currency.to_decimal(123)
        assert currency.from_decimal(amount) == 123, currency.from_decimal(amount)

        currency = asyncql.Currency({
            'name': 'Mean Minute/Hour',
            'prefix_symbol': 'Ħ',
            'fx_symbol': 'MMH',
            'base': 60,
            'unit_divisions': 2,
        })

        assert currency.format(60*60*1.23) == 'Ħ1.23', currency.format(60*60*1.23)
        assert currency.format(60*60 + 45, decimal_places=2) == 'Ħ1.01', \
            currency.format(60*60 + 45, decimal_places=2)
        assert currency.to_decimal(5011) == Decimal(5011)/Decimal(60*60)

        amount = (60**2)*2 + (60**1)*2 + (60**0)*3
        assert currency.get_units(amount) == (2, 2, 3)
        assert currency.format(amount, use_decimal=False, divider=':') == 'Ħ02:02:03', \
            currency.format(amount, use_decimal=False, divider=':')

        amount = Decimal('1.51')
        assert currency.from_decimal(amount) == 5436, currency.from_decimal(amount)

    def test_publish_migrations(self):
        assert len(os.listdir(MIGRATIONS_PATH)) < 2, os.listdir(MIGRATIONS_PATH)
        bookchain.publish_migrations(MIGRATIONS_PATH)
        assert len(os.listdir(MIGRATIONS_PATH)) > 2, os.listdir(MIGRATIONS_PATH)

    def test_automigrate(self):
        bookchain.set_connection_info(DB_FILEPATH)
        bookchain.publish_migrations(MIGRATIONS_PATH)
        with self.assertRaises(OperationalError):
            models.Account.query().count()
        bookchain.automigrate(MIGRATIONS_PATH, DB_FILEPATH)
        assert models.Account.query().count() == 0


if __name__ == '__main__':
    unittest.main()



from .Account import Account, AccountType
from .AccountCategory import AccountCategory
from .ArchivedEntry import ArchivedEntry
from .ArchivedTransaction import ArchivedTransaction
from .Correspondence import Correspondence
from .Currency import Currency
from .Customer import Customer
from .Entry import Entry, EntryType
from .Identity import Identity
from .Ledger import Ledger, LedgerType
from .Transaction import Transaction
from .TxRollup import TxRollup
from .Vendor import Vendor
from sqloquent.asyncql import (
    AsyncDeletedModel, AsyncAttachment,
    async_contains, async_within, async_has_many, async_belongs_to,
    async_has_one,
)


Identity.ledgers = async_has_many(Identity, Ledger, 'identity_id')
Ledger.owner = async_belongs_to(Ledger, Identity, 'identity_id')

Ledger.currency = async_belongs_to(Ledger, Currency, 'currency_id')
Currency.ledgers = async_has_many(Currency, Ledger, 'currency_id')

Correspondence.ledgers = async_contains(Correspondence, Ledger, 'ledger_ids')

Identity.correspondences = async_within(Identity, Correspondence, 'identity_ids')
Correspondence.identities = async_contains(Correspondence, Identity, 'identity_ids')

Ledger.accounts = async_has_many(Ledger, Account, 'ledger_id')
Account.ledger = async_belongs_to(Account, Ledger, 'ledger_id')

Account.children = async_has_many(Account, Account, 'parent_id')
Account.parent = async_belongs_to(Account, Account, 'parent_id')

Account.category = async_belongs_to(Account, AccountCategory, 'category_id')
AccountCategory.accounts = async_has_many(AccountCategory, Account, 'category_id')

Account.entries = async_has_many(Account, Entry, 'account_id')
Entry.account = async_belongs_to(Entry, Account, 'account_id')

Entry.transactions = async_within(Entry, Transaction, 'entry_ids')
Transaction.entries = async_contains(Transaction, Entry, 'entry_ids')

Transaction.ledgers = async_contains(Transaction, Ledger, 'ledger_ids')
Ledger.transactions = async_within(Ledger, Transaction, 'ledger_ids')

TxRollup.ledger = async_belongs_to(TxRollup, Ledger, 'ledger_id')
Ledger.rollups = async_within(Ledger, TxRollup, 'ledger_id')

TxRollup.transactions = async_contains(TxRollup, Transaction, 'tx_ids')
Transaction.rollups = async_within(Transaction, TxRollup, 'tx_ids')

TxRollup.parent = async_belongs_to(TxRollup, TxRollup, 'parent_id')
TxRollup.child = async_has_one(TxRollup, TxRollup, 'parent_id')

TxRollup.correspondence = async_belongs_to(TxRollup, Correspondence, 'correspondence_id')
Correspondence.rollups = async_within(Correspondence, TxRollup, 'correspondence_id')

ArchivedEntry.transactions = async_within(ArchivedEntry, ArchivedTransaction, 'entry_ids')
ArchivedTransaction.entries = async_contains(ArchivedTransaction, ArchivedEntry, 'entry_ids')

ArchivedEntry.account = async_belongs_to(ArchivedEntry, Account, 'account_id')
Account.archived_entries = async_has_many(Account, ArchivedEntry, 'account_id')

ArchivedTransaction.ledgers = async_contains(ArchivedTransaction, Ledger, 'ledger_ids')
Ledger.archived_transactions = async_within(Ledger, ArchivedTransaction, 'ledger_ids')


def set_connection_info(db_file_path: str):
    """Set the connection info for all models to use the specified
        sqlite3 database file path.
    """
    Account.connection_info = db_file_path
    AccountCategory.connection_info = db_file_path
    ArchivedEntry.connection_info = db_file_path
    ArchivedTransaction.connection_info = db_file_path
    Correspondence.connection_info = db_file_path
    Currency.connection_info = db_file_path
    Customer.connection_info = db_file_path
    Entry.connection_info = db_file_path
    Identity.connection_info = db_file_path
    Ledger.connection_info = db_file_path
    Transaction.connection_info = db_file_path
    TxRollup.connection_info = db_file_path
    Vendor.connection_info = db_file_path
    AsyncDeletedModel.connection_info = db_file_path
    AsyncAttachment.connection_info = db_file_path


# no longer needed
del async_contains
del async_belongs_to
del async_within
del async_has_many
del async_has_one



from asyncio import run
from context import asyncql
from genericpath import isfile
from sqlite3 import OperationalError
from sqloquent.asyncql import AsyncDeletedModel
from time import time
import os
import sqloquent.tools
import unittest


DB_FILEPATH = 'tests/test.db'
MIGRATIONS_PATH = 'tests/migrations'


class TestAsyncBasicE2E(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        asyncql.Identity.connection_info = DB_FILEPATH
        asyncql.Currency.connection_info = DB_FILEPATH
        asyncql.Ledger.connection_info = DB_FILEPATH
        asyncql.Account.connection_info = DB_FILEPATH
        asyncql.AccountCategory.connection_info = DB_FILEPATH
        asyncql.Entry.connection_info = DB_FILEPATH
        asyncql.Transaction.connection_info = DB_FILEPATH
        AsyncDeletedModel.connection_info = DB_FILEPATH
        super().setUpClass()

    def setUp(self):
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        super().setUp()

    def tearDown(self):
        for file in os.listdir(MIGRATIONS_PATH):
            if isfile(f'{MIGRATIONS_PATH}/{file}'):
                os.remove(f'{MIGRATIONS_PATH}/{file}')
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        super().tearDown()

    def automigrate(self):
        sqloquent.tools.publish_migrations(MIGRATIONS_PATH)
        tomigrate = [
            asyncql.Identity, asyncql.Currency, asyncql.Ledger,
            asyncql.Account, asyncql.AccountCategory, asyncql.Entry,
            asyncql.Transaction,
        ]
        for model in tomigrate:
            name = model.__name__
            m = sqloquent.tools.make_migration_from_model(model, name)
            with open(f'{MIGRATIONS_PATH}/create_{name}.py', 'w') as f:
                f.write(m)
        sqloquent.tools.automigrate(MIGRATIONS_PATH, DB_FILEPATH)

    def test_e2e(self):
        with self.assertRaises(OperationalError):
            run(asyncql.Account.query().count())
        self.automigrate()
        assert run(asyncql.Account.query().count()) == 0

        # test empty Transaction
        (asyncql.Transaction()).details
        (asyncql.Transaction()).auth_scripts
        assert not run(asyncql.Transaction().validate())

        # setup account categories
        equity_acct_cat = run(asyncql.AccountCategory.insert({
            'name': 'Equity',
            'ledger_type': asyncql.LedgerType.PRESENT,
            'destination': 'Balance Sheet',
        }))
        assert equity_acct_cat is not None
        assert run(asyncql.AccountCategory.find(equity_acct_cat.id)) is not None
        asset_acct_cat = run(asyncql.AccountCategory.insert({
            'name': 'Asset',
            'ledger_type': asyncql.LedgerType.PRESENT,
            'destination': 'Balance Sheet',
        }))
        liability_acct_cat = run(asyncql.AccountCategory.insert({
            'name': 'Liability',
            'ledger_type': asyncql.LedgerType.PRESENT,
            'destination': 'Balance Sheet',
        }))

        # setup identity, currency, ledger, and some accounts
        identity = run(asyncql.Identity.insert({'name': 'Test Man'}))
        currency = run(asyncql.Currency.insert({
            'name': 'US Dollar',
            'prefix_symbol': '$',
            'fx_symbol': 'USD',
            'base': 10,
            'unit_divisions': 2,
        }))
        ledger: asyncql.Ledger = run(asyncql.Ledger.insert({
            'name': 'General Ledger',
            'identity_id': identity.id,
            'currency_id': currency.id,
            'type': asyncql.LedgerType.PRESENT,
        }))
        equity_acct = run(asyncql.Account.insert({
            'name': 'General Equity',
            'type': asyncql.AccountType.EQUITY,
            'ledger_id': ledger.id,
            'category_id': equity_acct_cat.id,
        }))
        asset_acct = run(asyncql.Account.insert({
            'name': 'General Asset',
            'type': asyncql.AccountType.ASSET,
            'ledger_id': ledger.id,
            'category_id': asset_acct_cat.id,
        }))
        liability_acct = run(asyncql.Account.insert({
            'name': 'General Liability',
            'type': asyncql.AccountType.LIABILITY,
            'ledger_id': ledger.id,
            'category_id': liability_acct_cat.id,
        }))

        assert equity_acct.category.id == equity_acct_cat.id
        assert asset_acct.category.id == asset_acct_cat.id
        assert liability_acct.category.id == liability_acct_cat.id

        assert len(liability_acct_cat.accounts) == 1, liability_acct_cat.accounts

        # make sub account
        assert len(liability_acct.children) == 0
        liability_sub_acct = run(asyncql.Account.insert({
            'name': 'Liability Sub Account',
            'type': asyncql.AccountType.LIABILITY,
            'ledger_id': ledger.id,
            'parent_id': liability_acct.id,
            'category_id': liability_acct_cat.id,
        }))
        assert liability_sub_acct.parent is not None, liability_sub_acct.parent
        assert liability_sub_acct.parent.id == liability_acct.id, liability_sub_acct.parent
        run(liability_acct.children().reload())
        assert len(liability_acct.children) == 1
        assert liability_acct.children[0].id == liability_sub_acct.id

        # test empty Entry
        (asyncql.Entry()).details

        # prepare and save a valid transaction
        txn_nonce = os.urandom(16)
        equity_entry = asyncql.Entry({
            'type': asyncql.EntryType.CREDIT,
            'account_id': equity_acct.id,
            'amount': 10_000_00,
            'nonce': txn_nonce,
        })
        equity_entry.account = equity_acct
        asset_entry = asyncql.Entry({
            'type': asyncql.EntryType.DEBIT,
            'account_id': asset_acct.id,
            'amount': 10_000_00,
            'nonce': txn_nonce,
        })
        asset_entry.account = asset_acct
        txn = run(asyncql.Transaction.prepare(
            [equity_entry, asset_entry], str(time()),
            details='Starting capital asset'
        ))
        assert run(txn.validate())
        run(equity_entry.save())
        run(asset_entry.save())
        run(txn.save())
        # reload txn from database and validate it
        txn: asyncql.Transaction = run(asyncql.Transaction.find(txn.id))
        assert run(txn.validate(reload=True))

        # check balances
        assert run(equity_acct.balance()) == 10_000_00, run(equity_acct.balance())
        assert run(asset_acct.balance()) == 10_000_00, run(asset_acct.balance())
        assert run(liability_acct.balance()) == 0, run(liability_acct.balance())
        balances = run(ledger.balances(True))
        assert len(balances.keys()) == 4, balances
        assert balances[equity_acct.id][0] == 10_000_00, balances[equity_acct.id][0]
        assert balances[asset_acct.id][0] == 10_000_00, balances[asset_acct.id][0]
        assert balances[liability_acct.id][0] == 0, balances[liability_acct.id][0]
        assert balances[liability_sub_acct.id][0] == 0, balances[liability_sub_acct.id][0]

        # prepare and save valid transaction for liability sub account
        txn_nonce = os.urandom(16)
        equity_entry = run(asyncql.Entry.insert({
            'type': asyncql.EntryType.DEBIT,
            'account_id': equity_acct.id,
            'amount': 9_99,
            'nonce': txn_nonce,
        }))
        liability_entry = run(asyncql.Entry.insert({
            'type': asyncql.EntryType.CREDIT,
            'account_id': liability_sub_acct.id,
            'amount': 9_99,
            'nonce': txn_nonce,
        }))
        txn = run(asyncql.Transaction.prepare([equity_entry, liability_entry], str(time())))
        assert run(txn.validate())
        run(txn.save())

        # check balances
        assert run(equity_acct.balance()) == 10_000_00-9_99, run(equity_acct.balance())
        assert run(asset_acct.balance()) == 10_000_00, run(asset_acct.balance())
        assert run(liability_sub_acct.balance()) == 9_99, run(liability_acct.balance())
        assert run(liability_acct.balance()) == 9_99, run(liability_acct.balance())
        assert run(liability_acct.balance(False)) == 0, run(liability_acct.balance(False))
        balances = run(ledger.balances(True))
        assert len(balances.keys()) == 4, balances
        assert balances[equity_acct.id][0] == 10_000_00-9_99, balances[equity_acct.id][0]
        assert balances[asset_acct.id][0] == 10_000_00, balances[asset_acct.id][0]
        assert balances[liability_acct.id][0] == 0, balances[liability_acct.id][0]
        assert balances[liability_sub_acct.id][0] == 9_99, balances[liability_sub_acct.id][0]

        # prepare invalid transaction: reused entries
        with self.assertRaises(ValueError) as e:
            txn = run(asyncql.Transaction.prepare([equity_entry, asset_entry], str(int(time()))))
        assert 'already contained within a Transaction' in str(e.exception)

        # prepare invalid transaction: unbalanced entries
        txn_nonce = os.urandom(16)
        equity_entry = asyncql.Entry({
            'type': asyncql.EntryType.CREDIT,
            'account_id': equity_acct.id,
            'amount': 10_00,
            'nonce': txn_nonce,
        })
        equity_entry.account = equity_acct
        asset_entry = asyncql.Entry({
            'type': asyncql.EntryType.DEBIT,
            'account_id': asset_acct.id,
            'amount': 10_01,
            'nonce': txn_nonce,
        })
        asset_entry.account = asset_acct
        with self.assertRaises(ValueError) as e:
            txn = run(asyncql.Transaction.prepare([equity_entry, asset_entry], str(int(time()))))
        assert 'unbalanced' in str(e.exception)

        # delete something
        deleted = run(identity.delete())
        assert isinstance(deleted, AsyncDeletedModel)
        assert run(asyncql.Identity.find(identity.id)) is None

        # restore deleted identity
        restored = run(deleted.restore({'Identity': asyncql.Identity}))
        assert isinstance(restored, asyncql.Identity)
        assert restored.id == identity.id
        run(restored.save())
        assert run(asyncql.Identity.find(identity.id)) is not None


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from .Account import Account, AccountType
from .Entry import Entry, EntryType
from .Identity import Identity
from .Ledger import Ledger
from sqloquent.asyncql import AsyncHashedModel, AsyncRelatedCollection
from sqloquent.errors import vert, tert
import packify


class Correspondence(AsyncHashedModel):
    table: str = 'correspondences'
    id_column: str = 'id'
    columns: tuple[str] = ('id', 'identity_ids', 'ledger_ids', 'details', 'signatures')
    columns_excluded_from_hash: tuple[str] = ('signatures')
    id: str
    identity_ids: str
    ledger_ids: str
    details: bytes
    signatures: bytes|None
    identities: AsyncRelatedCollection
    ledgers: AsyncRelatedCollection
    rollups: AsyncRelatedCollection

    @property
    def details(self) -> dict:
        """Returns the details of the correspondence as a dict."""
        return packify.unpack(self.data.get('details', b'M@\x00'))
    @details.setter
    def details(self, val: dict):
        """Sets the details of the correspondence as a dict. Raises
            TypeError if the value is not a dict.
        """
        tert(type(val) is dict, 'details must be type dict')
        self.data['details'] = packify.pack(val)

    @property
    def signatures(self) -> dict[str, bytes]:
        """Returns the signatures of the correspondences as a dict
            mapping Identity ID to bytes signature.
        """
        return packify.unpack(self.data.get('signatures', b'M@\x00'))
    @signatures.setter
    def signatures(self, val: dict[str, bytes]):
        """Sets the signatures of the correspondences as a dict
            mapping IDs to bytes signature. Raises TypeError for invalid
            type or ValueError if an ID is not of the correspondence or
            its identities.
        """
        tert(type(val) is dict, 'signatures must be type dict[str, bytes]')
        for k, v in val.items():
            tert(type(k) is str, 'signatures must be type dict[str, bytes]')
            tert(type(v) is bytes, 'signatures must be type dict[str, bytes]')
            vert(k in self.identity_ids or k == self.id,
                 f'ID {k} not of correspondence or one of its identities')
        self.data['signatures'] = packify.pack(val)

    @property
    def txru_lock(self) -> bytes|None:
        """Returns the txru_lock directly from the details field."""
        return self.details.get('txru_lock', None)

    async def get_accounts(self) -> dict[str, dict[AccountType, Account]]:
        """Loads the relevant nostro and vostro Accounts for the
            Identities that are part of the Correspondence.
        """
        await self.ledgers().reload()
        await self.identities().reload()
        accounts = {
            identity.id: {}
            for identity in self.identities
        }
        for id1 in self.identities:
            id1: Identity
            for identity in [i for i in self.identities if i.id != id1.id]:
                identity: Identity
                for acct in await id1.get_correspondent_accounts(identity):
                    await acct.ledger().reload()
                    if acct.ledger.identity_id == identity.id:
                        accounts[identity.id][acct.type] = acct
        for ledger in self.ledgers:
            ledger: Ledger
            acct = await ledger.accounts().query().equal(
                type=AccountType.EQUITY.value
            ).first()
            if acct is not None:
                await acct.ledger().reload()
                accounts[acct.ledger.identity_id][acct.type] = acct
        return accounts

    async def setup_accounts(
            self, locking_scripts: dict[str, bytes]
        ) -> dict[str, dict[AccountType, Account]]:
        """Takes a dict mapping Identity ID to tapescript locking
            scripts. Returns a dict of Accounts necessary for setting up
            the credit Correspondence of form
            { identity.id: { AccountType: Account }}.
        """
        accounts = {}
        await self.identities().reload()
        for identity1 in self.identities:
            identity1: Identity
            for identity2 in [i for i in self.identities if i.id != identity1.id]:
                await identity1.ledgers().reload()
                ledger = [l for l in identity1.ledgers if l.id in self.ledger_ids][0]
                identity2: Identity
                nostro = Account({
                    'name': f'Receivable from (Nostro with) {identity2.name} ({identity2.id})',
                    'type': AccountType.NOSTRO_ASSET.value,
                    'ledger_id': ledger.id,
                })
                nostro.details = identity2.id
                nostro.locking_scripts = {
                    EntryType.CREDIT: locking_scripts[identity1.id],
                    EntryType.DEBIT: locking_scripts[identity2.id],
                }
                vostro = Account({
                    'name': f'Payable to (Vostro for) {identity2.name} ({identity2.id})',
                    'type': AccountType.VOSTRO_LIABILITY.value,
                    'ledger_id': ledger.id,
                })
                vostro.details = identity2.id
                vostro.locking_scripts = {
                    EntryType.CREDIT: locking_scripts[identity1.id],
                    EntryType.DEBIT: locking_scripts[identity2.id],
                }
                accounts[identity1.id] = {
                    AccountType.NOSTRO_ASSET: nostro,
                    AccountType.VOSTRO_LIABILITY: vostro,
                }
        return accounts

    async def pay_correspondent(
            self, payer: Identity, payee: Identity, amount: int, txn_nonce: bytes
        ) -> tuple[list[Entry], list[Entry]]:
        """Prepares two lists of entries in which the payer remits to
            the payee the given amount: one in which the nostro account
            on the payer's ledger is credited and one in which the
            vostro account on the payer's ledger is credited.
        """
        vert(payer.id in self.identity_ids,
             f'payer ({payer.name}, {payer.id}) not in correspondence identities')
        vert(payee.id in self.identity_ids,
             f'payee ({payee.name}, {payee.id}) not in correspondence identities')
        await payer.ledgers().reload()
        await payee.ledgers().reload()
        payer_ledger: Ledger = [l for l in payer.ledgers if l.id in self.ledger_ids][0]
        payee_ledger: Ledger = [l for l in payee.ledgers if l.id in self.ledger_ids][0]
        await payer_ledger.accounts().reload()
        await payee_ledger.accounts().reload()
        payer_equity_acct: Account = [
            a for a in payer_ledger.accounts if a.type is AccountType.EQUITY
        ][0]
        payee_equity_acct: Account = [
            a for a in payee_ledger.accounts if a.type is AccountType.EQUITY
        ][0]
        accts = await self.get_accounts()
        payer_nostro_acct = accts[payer.id][AccountType.NOSTRO_ASSET]
        payer_vostro_acct = accts[payer.id][AccountType.VOSTRO_LIABILITY]
        payee_nostro_acct = accts[payee.id][AccountType.NOSTRO_ASSET]
        payee_vostro_acct = accts[payee.id][AccountType.VOSTRO_LIABILITY]

        payer_equity_entry = Entry({
            'nonce': txn_nonce,
            'type': EntryType.DEBIT,
            'amount': amount,
            'account_id': payer_equity_acct.id,
        })
        payer_equity_entry.account = payer_equity_acct
        payer_nostro_entry = Entry({
            'nonce': txn_nonce,
            'type': EntryType.CREDIT,
            'amount': amount,
            'account_id': payer_nostro_acct.id,
        })
        payer_nostro_entry.account = payer_nostro_acct
        payer_vostro_entry = Entry({
            'nonce': txn_nonce,
            'type': EntryType.CREDIT,
            'amount': amount,
            'account_id': payer_vostro_acct.id,
        })
        payer_vostro_entry.account = payer_vostro_acct

        payee_equity_entry = Entry({
            'nonce': txn_nonce,
            'type': EntryType.CREDIT,
            'amount': amount,
            'account_id': payee_equity_acct.id,
        })
        payee_equity_entry.account = payee_equity_acct
        payee_nostro_entry = Entry({
            'nonce': txn_nonce,
            'type': EntryType.DEBIT,
            'amount': amount,
            'account_id': payee_nostro_acct.id,
        })
        payee_nostro_entry.account = payee_nostro_acct
        payee_vostro_entry = Entry({
            'nonce': txn_nonce,
            'type': EntryType.DEBIT,
            'amount': amount,
            'account_id': payee_vostro_acct.id,
        })
        payee_vostro_entry.account = payee_vostro_acct

        return (
            [payer_equity_entry, payee_equity_entry, payer_nostro_entry, payee_vostro_entry],
            [payer_equity_entry, payee_equity_entry, payer_vostro_entry, payee_nostro_entry],
        )

    async def balances(
            self, rolled_up_balances: dict[str, tuple[EntryType, int]] = {}
        ) -> dict[str, int]:
        """Returns the balances of the correspondents as a dict mapping
            str Identity ID to signed int (equal to Nostro - Vostro).
        """
        accts = await self.get_accounts()
        accts = [a for accts in accts.values() for a in accts.values()]
        balances = {}
        for acct in accts:
            await acct.ledger().reload()
            if acct.ledger.identity_id not in balances:
                balances[acct.ledger.identity_id] = 0
            if acct.type is AccountType.NOSTRO_ASSET:
                balances[acct.ledger.identity_id] += await acct.balance(
                    rolled_up_balances=rolled_up_balances
                )
            if acct.type is AccountType.VOSTRO_LIABILITY:
                balances[acct.ledger.identity_id] -= await acct.balance(
                    rolled_up_balances=rolled_up_balances
                )
        return balances



from decimal import Decimal
from sqloquent.asyncql import AsyncHashedModel, AsyncRelatedCollection


class Currency(AsyncHashedModel):
    connection_info: str = ''
    table: str = 'currencies'
    id_column: str = 'id'
    columns: tuple[str] = (
        'id', 'name', 'prefix_symbol', 'postfix_symbol',
        'fx_symbol', 'unit_divisions', 'base', 'details'
    )
    id: str
    name: str
    prefix_symbol: str|None
    postfix_symbol: str|None
    fx_symbol: str|None
    unit_divisions: int
    base: int|None
    details: str|None
    ledgers: AsyncRelatedCollection

    def to_decimal(self, amount: int) -> Decimal:
        """Convert the amount into a Decimal representation."""
        base = self.base or 10
        return Decimal(amount) / Decimal(base**self.unit_divisions)

    def from_decimal(self, amount: Decimal) -> int:
        """Convert the amount from a Decimal representation."""
        base = self.base or 10
        return int(amount * base**self.unit_divisions)

    def get_units(self, amount: int) -> tuple[int,]:
        """Get the full units and subunits. The number of subunit
            figures will be equal to unit_divisions; e.g. if base=10
            and unit_divisions=2, get_units(200) will return (2, 0, 0);
            if base=60 and unit_divisions=2, get_units(200) will return
            (0, 3, 20).
        """
        def get_subunits(amount, base, unit_divisions):
            units_and_change = divmod(amount, base ** unit_divisions)
            if unit_divisions > 1:
                units_and_change = (
                    units_and_change[0],
                    *get_subunits(units_and_change[1], base, unit_divisions-1)
                )
            return units_and_change
        base = self.base or 10
        unit_divisions = self.unit_divisions
        return get_subunits(amount, base, unit_divisions)

    def format(
            self, amount: int, *,
            use_decimal: bool = True, decimal_places: int = 2,
            use_prefix: bool = True, use_postfix: bool = False,
            use_fx_symbol: bool = False, divider: str = '.'
        ) -> str:
        """Format an amount using the correct number of `decimal_places`.
            If `use_decimal` is `False`, instead the unit subdivisions
            from `get_units` will be combined using the `divider`
            char, and each part will be prefix padded with 0s to reach
            the `decimal_places`. E.g. `.format(200, use_decimal=False,
            divider=':') == '02:00'` for a Currency with `base=100` and
            `unit_divisions=1`.
        """
        if use_decimal:
            amount: str = str(self.to_decimal(amount))
            if '.' not in amount:
                amount += '.'
            digits = amount.split('.')[1]

            while len(digits) < decimal_places:
                digits += '0'

            digits = digits[:decimal_places]
            amount = f"{amount.split('.')[0]}.{digits}"
        else:
            units = self.get_units(amount)
            amount = ''
            for u in units:
                p = str(u)
                while len(p) < decimal_places:
                    p = "0" + p
                amount = f"{amount}:{p}"
            amount = amount[1:]

        if self.postfix_symbol and use_postfix:
            return f"{amount}{self.postfix_symbol}"

        if self.fx_symbol and use_fx_symbol:
            return f"{amount} {self.fx_symbol}"

        if self.prefix_symbol and use_prefix:
            return f"{self.prefix_symbol}{amount}"

        return amount




from context import models
from genericpath import isfile
from nacl.signing import SigningKey
from packify import pack
from time import time
import os
import sqloquent.tools
import tapescript
import unittest


DB_FILEPATH = 'tests/test.db'
MIGRATIONS_PATH = 'tests/migrations'
MODELS_PATH = 'bookchain/models'


class TestCorrespondencesE2E(unittest.TestCase):
    @classmethod
    def automigrate(cls):
        sqloquent.tools.publish_migrations(MIGRATIONS_PATH)
        tomigrate = [
            models.Identity, models.Currency, models.Ledger,
            models.Account, models.Entry, models.Transaction,
            models.Correspondence,
        ]
        for model in tomigrate:
            name = model.__name__
            m = sqloquent.tools.make_migration_from_model(model, name)
            with open(f'{MIGRATIONS_PATH}/create_{name}.py', 'w') as f:
                f.write(m)
        sqloquent.tools.automigrate(MIGRATIONS_PATH, DB_FILEPATH)

    @classmethod
    def setUpClass(cls):
        models.Identity.connection_info = DB_FILEPATH
        models.Currency.connection_info = DB_FILEPATH
        models.Correspondence.connection_info = DB_FILEPATH
        models.Ledger.connection_info = DB_FILEPATH
        models.Account.connection_info = DB_FILEPATH
        models.Entry.connection_info = DB_FILEPATH
        models.Transaction.connection_info = DB_FILEPATH
        sqloquent.DeletedModel.connection_info = DB_FILEPATH
        cls.automigrate()
        super().setUpClass()

    def setUp(self):
        models.Identity.query().delete()
        models.Currency.query().delete()
        models.Correspondence.query().delete()
        models.Ledger.query().delete()
        models.Account.query().delete()
        models.Entry.query().delete()
        models.Transaction.query().delete()
        sqloquent.DeletedModel.query().delete()
        self.setup_cryptographic_values()
        super().setUp()

    def tearDown(self):
        for file in os.listdir(MIGRATIONS_PATH):
            if isfile(f'{MIGRATIONS_PATH}/{file}'):
                os.remove(f'{MIGRATIONS_PATH}/{file}')
        super().tearDown()

    @classmethod
    def tearDownClass(cls):
        if isfile(DB_FILEPATH):
            os.remove(DB_FILEPATH)
        return super().tearDownClass()

    def setup_cryptographic_values(self):
        self.seed_alice = os.urandom(32)
        self.seed_bob = os.urandom(32)
        self.seed_charlie = os.urandom(32)
        self.pkey_alice = bytes(SigningKey(self.seed_alice).verify_key)
        self.pkey_bob = bytes(SigningKey(self.seed_bob).verify_key)
        self.pkey_charlie = bytes(SigningKey(self.seed_charlie).verify_key)
        self.committed_script_alice = tapescript.tools.make_delegate_key_lock(self.pkey_alice)
        self.committed_script_bob = tapescript.tools.make_delegate_key_lock(self.pkey_bob)
        self.locking_script_alice = tapescript.tools.make_taproot_lock(
            self.pkey_alice,
            self.committed_script_alice
        ).bytes
        self.locking_script_bob = tapescript.tools.make_taproot_lock(
            self.pkey_bob,
            self.committed_script_bob
        ).bytes
        self.delegate_seed = os.urandom(32) # for simplicity, both will have the same delegate
        self.delegate_pkey = bytes(SigningKey(self.delegate_seed).verify_key)
        begin_ts = int(time()) - 1
        end_ts = int(time()) + 60*60*24*365
        self.delegate_cert_alice = tapescript.make_delegate_key_cert(
            self.seed_alice, self.delegate_pkey, begin_ts, end_ts
        )
        self.delegate_cert_bob = tapescript.make_delegate_key_cert(
            self.seed_bob, self.delegate_pkey, begin_ts, end_ts
        )
        self.multisig_lock = tapescript.tools.make_multisig_lock(
            [self.pkey_alice, self.pkey_bob], 2
        ).bytes

    def test_e2e(self):
        assert models.Account.query().count() == 0

        # set up currency
        currency = models.Currency.insert({
            'name': 'Median Human Hour',
            'prefix_symbol': 'Ħ',
            'fx_symbol': 'MHH',
            'base': 60,
            'unit_divisions': 2,
            'details': 'Abstract value of one median hour of human time. ' +
                '1 Hour = 60 Minutes = 3600 Seconds',
        })

        # set up alice, ledger_alice, accounts, and starting capital transaction
        alice: models.Identity = models.Identity.insert({
            'name': 'Alice',
            'pubkey': self.pkey_alice,
            'seed': self.seed_alice,
        })
        ledger_alice = models.Ledger.insert({
            'name': 'Current Ledger',
            'identity_id': alice.id,
            'currency_id': currency.id,
        })
        equity_acct_alice = models.Account({
            'name': 'General Equity (Alice)',
            'type': models.AccountType.EQUITY.value,
            'ledger_id': ledger_alice.id,
        })
        equity_acct_alice.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_alice
        }
        equity_acct_alice.save()
        asset_acct_alice = models.Account({
            'name': 'General Asset (Alice)',
            'type': models.AccountType.ASSET.value,
            'ledger_id': ledger_alice.id,
        })
        asset_acct_alice.locking_scripts = {
            models.EntryType.CREDIT: self.locking_script_alice
        }
        asset_acct_alice.save()
        liability_acct_alice = models.Account.insert({
            'name': 'General Liability (Alice)',
            'type': models.AccountType.LIABILITY.value,
            'ledger_id': ledger_alice.id,
        })
        liability_acct_alice.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_alice,
            models.EntryType.CREDIT: self.locking_script_alice,
        }

        # fund the Identity with starting capital
        nonce = os.urandom(16)
        equity_entry = models.Entry({
            'type': models.EntryType.CREDIT,
            'account_id': equity_acct_alice.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        equity_entry.account = equity_acct_alice
        asset_entry = models.Entry({
            'type': models.EntryType.DEBIT,
            'account_id': asset_acct_alice.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        asset_entry.account = asset_acct_alice
        txn = models.Transaction.prepare([equity_entry, asset_entry], str(time()))
        equity_entry.save()
        asset_entry.save()
        txn.save()

        # set up bob, ledger_bob, and some accounts
        bob: models.Identity = models.Identity.insert({
            'name': 'Bob',
            'pubkey': self.pkey_bob,
            'seed': self.seed_bob,
        })
        ledger_bob = models.Ledger.insert({
            'name': 'Current Ledger',
            'identity_id': bob.id,
            'currency_id': currency.id,
        })
        equity_acct_bob = models.Account({
            'name': 'General Equity (Bob)',
            'type': models.AccountType.EQUITY.value,
            'ledger_id': ledger_bob.id,
        })
        equity_acct_bob.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_bob
        }
        equity_acct_bob.save()
        asset_acct_bob = models.Account({
            'name': 'General Asset (Bob)',
            'type': models.AccountType.ASSET.value,
            'ledger_id': ledger_bob.id,
        })
        asset_acct_bob.locking_scripts = {
            models.EntryType.CREDIT: self.locking_script_bob
        }
        asset_acct_bob.save()
        liability_acct_bob = models.Account.insert({
            'name': 'General Liability (Bob)',
            'type': models.AccountType.LIABILITY.value,
            'ledger_id': ledger_bob.id,
        })
        liability_acct_bob.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_bob,
            models.EntryType.CREDIT: self.locking_script_bob,
        }
        liability_acct_bob.save()

        # fund the Identity with starting capital
        nonce = os.urandom(16)
        equity_entry = models.Entry({
            'type': models.EntryType.CREDIT,
            'account_id': equity_acct_bob.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        equity_entry.account = equity_acct_bob
        asset_entry = models.Entry({
            'type': models.EntryType.DEBIT,
            'account_id': asset_acct_bob.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        asset_entry.account = asset_acct_bob
        txn = models.Transaction.prepare([equity_entry, asset_entry], str(time()))
        equity_entry.save()
        asset_entry.save()
        txn.save()

        # charlie exists but shouldn't affect anything else
        charlie: models.Identity = models.Identity.insert({
            'name': 'Charlie',
            'pubkey': self.pkey_charlie,
        })
        ledger_charlie = models.Ledger.insert({
            'name': 'Current Ledger',
            'identity_id': charlie.id,
            'currency_id': currency.id,
        })

        # test empty Correspondence
        (models.Correspondence()).details
        (models.Correspondence()).signatures
        (models.Correspondence()).txru_lock

        # set up Correspondence
        assert alice.correspondences().query().count() == 0
        assert len(alice.correspondents(reload=True)) == 0
        correspondence: models.Correspondence = models.Correspondence.insert({
            'identity_ids': ','.join(sorted([alice.id, bob.id])),
            'ledger_ids': ','.join(sorted([ledger_alice.id, ledger_bob.id])),
            'details': pack({
                'starting_capital': 'Ħ2000',
                'limits': {
                    alice.id: 'Ħ166 M40',
                    bob.id: 'Ħ166 M40',
                },
                'locking_scripts': {
                    alice.id: self.locking_script_alice,
                    bob.id: self.locking_script_bob,
                },
                'txru_lock': self.multisig_lock,
            })
        })
        alice_sig = tapescript.make_single_sig_witness(
            alice.seed, {'sigfield1': bytes.fromhex(correspondence.id)}
        )
        bob_sig = tapescript.make_single_sig_witness(
            bob.seed, {'sigfield1': bytes.fromhex(correspondence.id)}
        )
        correspondence.signatures = {
            alice.id: alice_sig.bytes,
            bob.id: bob_sig.bytes,
            correspondence.id: (alice_sig + bob_sig).bytes,
        }
        correspondence.save()
        assert alice.correspondences().query().count() == 1
        assert len(alice.correspondents(reload=True)) == 1
        assert alice.correspondents()[0].id == bob.id
        assert bob.correspondences().query().count() == 1
        assert len(bob.correspondents(reload=True)) == 1
        assert bob.correspondents()[0].id == alice.id
        assert len(correspondence.ledgers)

        # set up correspondent accounts for alice
        cor_accts1 = alice.get_correspondent_accounts(bob)
        assert len(cor_accts1) == 0, cor_accts1
        nostro_alice = models.Account()
        nostro_alice.name = f'Receivable from {bob.name} ({bob.id})'
        nostro_alice.type = models.AccountType.NOSTRO_ASSET
        nostro_alice.ledger_id = ledger_alice.id
        nostro_alice.details = bob.id
        nostro_alice.locking_scripts = {
            models.EntryType.CREDIT: self.locking_script_alice,
            models.EntryType.DEBIT: self.locking_script_bob,
        }
        nostro_alice.save()

        vostro_alice = models.Account()
        vostro_alice.name = f'Payable to {bob.name} ({bob.id})'
        vostro_alice.type = models.AccountType.VOSTRO_LIABILITY
        vostro_alice.ledger_id = ledger_alice.id
        vostro_alice.details = bob.id
        vostro_alice.locking_scripts = {
            models.EntryType.CREDIT: self.locking_script_alice,
            models.EntryType.DEBIT: self.locking_script_bob,
        }
        vostro_alice.save()
        cor_accts1 = alice.get_correspondent_accounts(bob)
        assert len(cor_accts1) == 2, cor_accts1

        # set up correspondent accounts for bob
        cor_accts2 = bob.get_correspondent_accounts(alice)
        assert len(cor_accts2) == 2, cor_accts2

        nostro_bob = models.Account()
        nostro_bob.name = f'Receivable from {alice.name} ({alice.id})'
        nostro_bob.type = models.AccountType.NOSTRO_ASSET
        nostro_bob.ledger_id = ledger_bob.id
        nostro_bob.details = alice.id
        nostro_bob.locking_scripts = {
            models.EntryType.CREDIT: self.locking_script_bob,
            models.EntryType.DEBIT: self.locking_script_alice,
        }
        nostro_bob.save()

        vostro_bob = models.Account()
        vostro_bob.name = f'Payable to {alice.name} ({alice.id})'
        vostro_bob.type = models.AccountType.VOSTRO_LIABILITY
        vostro_bob.ledger_id = ledger_bob.id
        vostro_bob.details = alice.id
        vostro_bob.locking_scripts = {
            models.EntryType.CREDIT: self.locking_script_bob,
            models.EntryType.DEBIT: self.locking_script_alice,
        }
        vostro_bob.save()
        cor_accts2 = bob.get_correspondent_accounts(alice)
        assert len(cor_accts2) == 4, cor_accts2
        for acct in cor_accts2:
            print(f"{acct.ledger_id} {acct.name} {acct.type.name}")

        # get correspondence accounts
        cor_accts = correspondence.get_accounts()
        assert alice.id in cor_accts
        assert bob.id in cor_accts
        assert models.AccountType.NOSTRO_ASSET in cor_accts[alice.id]
        assert models.AccountType.VOSTRO_LIABILITY in cor_accts[alice.id]
        assert models.AccountType.EQUITY in cor_accts[alice.id]
        assert models.AccountType.NOSTRO_ASSET in cor_accts[bob.id]
        assert models.AccountType.VOSTRO_LIABILITY in cor_accts[bob.id]
        assert models.AccountType.EQUITY in cor_accts[bob.id]

        # create a valid payment transaction: Alice pays Bob 200
        nonce = os.urandom(16)
        equity_entry_alice = models.Entry({
            'type': models.EntryType.DEBIT,
            'amount': 200,
            'nonce': nonce,
            'account_id': equity_acct_alice.id,
        })
        equity_entry_alice.details = 'Debit Alice Equity'
        equity_entry_alice.account = equity_acct_alice
        liability_entry_alice = models.Entry({
            'type': models.EntryType.CREDIT,
            'amount': 200,
            'nonce': nonce,
            'account_id': vostro_alice.id,
        })
        liability_entry_alice.details = 'Credit Alice liability'
        liability_entry_alice.account = vostro_alice
        equity_entry_bob = models.Entry({
            'type': models.EntryType.CREDIT,
            'amount': 200,
            'nonce': nonce,
            'account_id': equity_acct_bob.id,
        })
        equity_entry_bob.details = 'Credit Bob Equity'
        equity_entry_bob.account = equity_acct_bob
        asset_entry_bob = models.Entry({
            'type': models.EntryType.DEBIT,
            'amount': 200,
            'nonce': nonce,
            'account_id': nostro_bob.id,
        })
        asset_entry_bob.details = 'Debit Bob Asset'
        asset_entry_bob.account = nostro_bob
        auth_scripts = {
            equity_acct_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                alice.seed, equity_entry_alice.get_sigfields(), self.committed_script_alice
            ).bytes,
            vostro_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                alice.seed, liability_entry_alice.get_sigfields(), self.committed_script_alice
            ).bytes,
            nostro_bob.id: tapescript.tools.make_taproot_witness_keyspend(
                alice.seed, asset_entry_bob.get_sigfields(), self.committed_script_alice
            ).bytes,
        }
        txn = models.Transaction.prepare(
            [equity_entry_alice, liability_entry_alice, equity_entry_bob, asset_entry_bob],
            str(time()), auth_scripts
        )
        txn.save()

        balances = correspondence.balances()
        assert balances[alice.id] == -balances[bob.id]

        # create an invalid transaction: valid auth, invalid entries
        nonce = os.urandom(16)
        equity_entry_alice = models.Entry({
            'type': models.EntryType.DEBIT,
            'amount': 100,
            'nonce': nonce,
            'account_id': equity_acct_alice.id,
        })
        equity_entry_alice.details = 'Debit Alice Equity'
        equity_entry_alice.account = equity_acct_alice
        liability_entry_alice = models.Entry({
            'type': models.EntryType.CREDIT,
            'amount': 100,
            'nonce': nonce,
            'account_id': vostro_alice.id,
        })
        liability_entry_alice.details = 'Credit Alice liability'
        liability_entry_alice.account = vostro_alice
        equity_entry_bob = models.Entry({
            'type': models.EntryType.CREDIT,
            'amount': 100,
            'nonce': nonce,
            'account_id': equity_acct_bob.id,
        })
        equity_entry_bob.details = 'Credit Bob Equity'
        equity_entry_bob.account = equity_acct_bob
        liability_entry_bob = models.Entry({
            'type': models.EntryType.DEBIT,
            'amount': 100,
            'nonce': nonce,
            'account_id': vostro_bob.id,
        })
        liability_entry_bob.details = 'Debit Bob Liability'
        liability_entry_bob.account = vostro_bob
        auth_scripts = {
            equity_acct_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                alice.seed, equity_entry_alice.get_sigfields(), self.committed_script_alice
            ).bytes,
            vostro_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                alice.seed, liability_entry_alice.get_sigfields(), self.committed_script_alice
            ).bytes,
            vostro_bob.id: tapescript.tools.make_taproot_witness_keyspend(
                alice.seed, liability_entry_bob.get_sigfields(), self.committed_script_alice
            ).bytes,
        }
        with self.assertRaises(AssertionError) as e:
            txn = models.Transaction.prepare(
                [equity_entry_alice, liability_entry_alice, equity_entry_bob, liability_entry_bob],
                str(time()), auth_scripts
            )
        assert 'validation failed' in str(e.exception)

    def test_helpers_e2e(self):
        assert models.Account.query().count() == 0

        # set up currency
        currency = models.Currency.insert({
            'name': 'Median Human Hour',
            'prefix_symbol': 'Ħ',
            'fx_symbol': 'MHH',
            'base': 60,
            'unit_divisions': 2,
            'details': 'Abstract value of one median hour of human time. ' +
                '1 Hour = 60 Minutes = 3600 Seconds',
        })

        # set up alice, ledger_alice, accounts, and starting capital transaction
        alice: models.Identity = models.Identity.insert({
            'name': 'Alice',
            'pubkey': self.pkey_alice,
            'seed': self.seed_alice,
        })
        ledger_alice = models.Ledger.insert({
            'name': 'Current Ledger',
            'identity_id': alice.id,
            'currency_id': currency.id,
        })
        ledger_alice.owner = alice
        asset_acct_alice, liability_acct_alice, equity_acct_alice = ledger_alice.setup_basic_accounts()

        equity_acct_alice.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_alice
        }
        equity_acct_alice.save()

        asset_acct_alice.locking_scripts = {
            models.EntryType.CREDIT: self.locking_script_alice
        }
        asset_acct_alice.save()

        liability_acct_alice.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_alice,
            models.EntryType.CREDIT: self.locking_script_alice,
        }
        liability_acct_alice.save()

        # fund the Identity with starting capital
        nonce = os.urandom(16)
        equity_entry = models.Entry({
            'type': models.EntryType.CREDIT,
            'account_id': equity_acct_alice.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        equity_entry.account = equity_acct_alice
        asset_entry = models.Entry({
            'type': models.EntryType.DEBIT,
            'account_id': asset_acct_alice.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        asset_entry.account = asset_acct_alice
        txn = models.Transaction.prepare([equity_entry, asset_entry], str(time()))
        equity_entry.save()
        asset_entry.save()
        txn.save()

        # set up bob, ledger_bob, and some accounts
        bob: models.Identity = models.Identity.insert({
            'name': 'Bob',
            'pubkey': self.pkey_bob,
            'seed': self.seed_bob,
        })
        ledger_bob = models.Ledger.insert({
            'name': 'Current Ledger',
            'identity_id': bob.id,
            'currency_id': currency.id,
        })
        asset_acct_bob, liability_acct_bob, equity_acct_bob = ledger_bob.setup_basic_accounts()
        equity_acct_bob.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_bob
        }
        equity_acct_bob.save()
        asset_acct_bob.locking_scripts = {
            models.EntryType.CREDIT: self.locking_script_bob
        }
        asset_acct_bob.save()
        liability_acct_bob.locking_scripts = {
            models.EntryType.DEBIT: self.locking_script_bob,
            models.EntryType.CREDIT: self.locking_script_bob,
        }
        liability_acct_bob.save()

        # fund the Identity with starting capital
        nonce = os.urandom(16)
        equity_entry = models.Entry({
            'type': models.EntryType.CREDIT,
            'account_id': equity_acct_bob.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        equity_entry.account = equity_acct_bob
        asset_entry = models.Entry({
            'type': models.EntryType.DEBIT,
            'account_id': asset_acct_bob.id,
            'amount': 2_000 * 60*60, # 2000 hours = 50 weeks * 40 hours/week
            'nonce': nonce,
        })
        asset_entry.account = asset_acct_bob
        txn = models.Transaction.prepare([equity_entry, asset_entry], str(time()))
        equity_entry.save()
        asset_entry.save()
        txn.save()

        # set up Correspondence
        assert alice.correspondences().query().count() == 0
        assert len(alice.correspondents(reload=True)) == 0
        correspondence: models.Correspondence = models.Correspondence.insert({
            'identity_ids': ','.join(sorted([alice.id, bob.id])),
            'ledger_ids': ','.join(sorted([ledger_alice.id, ledger_bob.id])),
            'details': pack({
                'starting_capital': 'Ħ2000',
                'limits': {
                    alice.id: 'Ħ166 M40',
                    bob.id: 'Ħ166 M40',
                },
                'locking_scripts': {
                    alice.id: self.locking_script_alice,
                    bob.id: self.locking_script_bob,
                },
                'txru_lock': self.multisig_lock,
            })
        })
        alice_sig = tapescript.make_single_sig_witness(
            alice.seed, {'sigfield1': bytes.fromhex(correspondence.id)}
        )
        bob_sig = tapescript.make_single_sig_witness(
            bob.seed, {'sigfield1': bytes.fromhex(correspondence.id)}
        )
        correspondence.signatures = {
            alice.id: alice_sig.bytes,
            bob.id: bob_sig.bytes,
            correspondence.id: (alice_sig + bob_sig).bytes,
        }
        correspondence.save()
        assert alice.correspondences().query().count() == 1
        assert len(alice.correspondents(reload=True)) == 1
        assert alice.correspondents()[0].id == bob.id
        assert bob.correspondences().query().count() == 1
        assert len(bob.correspondents(reload=True)) == 1
        assert bob.correspondents()[0].id == alice.id

        # set up correspondence accounts
        cor_accts = correspondence.setup_accounts({
            alice.id: self.locking_script_alice,
            bob.id: self.locking_script_bob,
        })
        for _, acct in cor_accts[alice.id].items():
            acct.save()
        for _, acct in cor_accts[bob.id].items():
            acct.save()
        nostro_acct_alice = cor_accts[alice.id][models.AccountType.NOSTRO_ASSET]
        vostro_acct_alice = cor_accts[alice.id][models.AccountType.VOSTRO_LIABILITY]
        nostro_acct_bob = cor_accts[bob.id][models.AccountType.NOSTRO_ASSET]
        vostro_acct_bob = cor_accts[bob.id][models.AccountType.VOSTRO_LIABILITY]

        # create txn: Alice credits her ledger's nostro to pay Bob
        entries, _ = correspondence.pay_correspondent(alice, bob, 100, os.urandom(16))
        assert len([e for e in entries if e.account_id == equity_acct_alice.id]) == 1
        assert len([e for e in entries if e.account_id == equity_acct_bob.id]) == 1
        assert len([e for e in entries if e.account_id == nostro_acct_alice.id]) == 1
        assert len([e for e in entries if e.account_id == vostro_acct_bob.id]) == 1
        txn = models.Transaction.prepare(entries, str(time()), auth_scripts={
            equity_acct_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_alice, entries[0].get_sigfields(), self.committed_script_alice
            ).bytes,
            nostro_acct_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_alice, entries[2].get_sigfields(), self.committed_script_alice
            ).bytes,
            vostro_acct_bob.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_alice, entries[3].get_sigfields(), self.committed_script_alice
            ).bytes,
        })
        assert models.Entry.query({'id': entries[0].id}).count() == 0
        txn.save()
        assert models.Entry.query({'id': entries[0].id}).count() == 1

        # create txn: Alice credits her ledger's vostro to pay Bob
        _, entries = correspondence.pay_correspondent(alice, bob, 100, os.urandom(16))
        assert len([e for e in entries if e.account_id == equity_acct_alice.id]) == 1
        assert len([e for e in entries if e.account_id == equity_acct_bob.id]) == 1
        assert len([e for e in entries if e.account_id == vostro_acct_alice.id]) == 1
        assert len([e for e in entries if e.account_id == nostro_acct_bob.id]) == 1
        txn = models.Transaction.prepare(entries, str(time()), auth_scripts={
            equity_acct_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_alice, entries[0].get_sigfields(), self.committed_script_alice
            ).bytes,
            vostro_acct_alice.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_alice, entries[2].get_sigfields(), self.committed_script_alice
            ).bytes,
            nostro_acct_bob.id: tapescript.tools.make_taproot_witness_keyspend(
                self.seed_alice, entries[3].get_sigfields(), self.committed_script_alice
            ).bytes,
        })
        txn.save()

        # check balances
        balances = correspondence.balances()
        assert type(balances) is dict
        assert alice.id in balances
        assert bob.id in balances
        assert balances[alice.id] == -200, balances
        assert balances[bob.id] == 200, balances


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from sqloquent.asyncql import AsyncHashedModel, AsyncRelatedCollection
from sqloquent.errors import vert, tert
from .Account import Account, AccountType
from .ArchivedTransaction import ArchivedTransaction
from .Correspondence import Correspondence
from .Entry import Entry, EntryType
from .Identity import Identity
import packify


class Transaction(AsyncHashedModel):
    """A Transaction is a collection of connected Entries that are
        recorded on the Ledgers of the Identities that are party to the
        Transaction. Any Entry for an Account that has a locking_script
        will require a valid tapscript unlocking script to be recorded
        in the auth_scripts dict of the Transaction.
    """
    connection_info: str = ''
    table: str = 'transactions'
    id_column: str = 'id'
    columns: tuple[str] = ('id', 'entry_ids', 'ledger_ids', 'timestamp', 'details', 'auth_scripts')
    columns_excluded_from_hash: tuple[str] = ('auth_scripts',)
    id: str
    entry_ids: str
    ledger_ids: str
    timestamp: str
    details: bytes
    auth_scripts: bytes
    entries: AsyncRelatedCollection
    ledgers: AsyncRelatedCollection
    rollups: AsyncRelatedCollection

    # override automatic properties
    @property
    def details(self) -> dict[str, bytes]:
        """A packify.SerializableType stored in the database as a blob."""
        return packify.unpack(self.data.get('details', b'M@\x00'))
    @details.setter
    def details(self, val: dict[str, bytes]):
        if type(val) is not dict:
            return
        if not all([type(k) is str and type(v) is bytes for k, v in val.items()]):
            return
        self.data['details'] = packify.pack(val)

    @property
    def auth_scripts(self) -> dict[str, bytes]:
        """A dict mapping account IDs to tapescript unlocking script bytes."""
        return packify.unpack(self.data.get('auth_scripts', b'M@\x00'))
    @auth_scripts.setter
    def auth_scripts(self, val: dict[str, bytes]):
        if type(val) is not dict:
            return
        if not all([type(k) is str and type(v) is bytes for k, v in val.items()]):
            return
        self.data['auth_scripts'] = packify.pack(val)

    @classmethod
    def _encode(cls, data: dict|None) -> dict|None:
        """Encode values for saving."""
        if type(data) is not dict:
            return None
        if type(data.get('auth_scripts', {})) is dict:
            data['auth_scripts'] = packify.pack(data.get('auth_scripts', {}))
        if type(data.get('details', {})) is dict:
            data['details'] = packify.pack(data.get('details', {}))
        return data

    @classmethod
    async def prepare(cls, entries: list[Entry], timestamp: str, auth_scripts: dict = {},
                details: packify.SerializableType = None,
                tapescript_runtime: dict = {}, reload: bool = False) -> Transaction:
        """Prepare a transaction. Raises TypeError for invalid arguments.
            Raises ValueError if the entries do not balance for each
            ledger; if a required auth script is missing; or if any of
            the entries is contained within an existing Transaction.
            Entries and Transaction will have IDs generated but will not
            be persisted to the database and must be saved separately.
            The auth_scripts dict must map account IDs to tapescript
            bytecode bytes.
        """
        tert(type(entries) is list and all([type(e) is Entry for e in entries]),
            'entries must be list[Entry]')
        tert(type(timestamp) is str, 'timestamp must be str')

        ledgers = set()
        for entry in entries:
            entry.id = entry.generate_id(entry.data)
            if reload:
                await entry.account().reload()
            vert(await Transaction.query().contains('entry_ids', entry.id).count() == 0,
                 f"entry {entry.id} is already contained within a Transaction")
            ledgers.add(entry.account.ledger_id)

        txn = cls({
            'entry_ids': ",".join(sorted([
                e.id if e.id else e.generate_id(e.data)
                for e in entries
            ])),
            'ledger_ids': ",".join(sorted(list(ledgers))),
            'timestamp': timestamp,
        })
        txn.auth_scripts = auth_scripts
        txn.details = details
        txn.entries = entries
        assert await txn.validate(tapescript_runtime, reload), \
            'transaction validation failed'
        txn.id = txn.generate_id(txn.data)
        return txn

    async def validate(self, tapescript_runtime: dict = {}, reload: bool = False) -> bool:
        """Determines if a Transaction is valid using the rules of accounting
            and checking all auth scripts against their locking scripts. The
            tapescript_runtime can be scoped to each entry ID. Raises TypeError
            for invalid arguments. Raises ValueError if the entries do not
            balance for each ledger; if a required auth script is missing; or
            if any of the entries is contained within an existing Transaction.
            If reload is set to True, entries and accounts will be reloaded
            from the database.
        """
        tert(type(self.auth_scripts) is dict,
            'auth_scripts must be dict mapping account ID to authorizing tapescript bytecode')

        if reload:
            await self.entries().reload()

        if len(self.entries) == 0:
            return False

        # first check that all ledgers balance
        ledgers = {}
        entry: Entry
        for entry in self.entries:
            vert(entry.account_id in self.auth_scripts or not entry.account.locking_scripts
                 or entry.type not in entry.account.locking_scripts,
                f"missing auth script for account {entry.account_id} ({entry.account.name})")
            if reload:
                await entry.account().reload()
            if entry.account.ledger_id not in ledgers:
                ledgers[entry.account.ledger_id] = {'Dr': 0, 'Cr': 0}
            if entry.type in (EntryType.CREDIT, EntryType.CREDIT.value):
                ledgers[entry.account.ledger_id]['Cr'] += entry.amount
            else:
                ledgers[entry.account.ledger_id]['Dr'] += entry.amount

        for ledger_id, balances in ledgers.items():
            vert(balances['Cr'] == balances['Dr'],
                f"ledger {ledger_id} unbalanced: {balances['Cr']} Cr != {balances['Dr']} Dr")

        # next check that all necessary authorizations are provided
        for entry in self.entries:
            acct = entry.account

            if not acct.locking_scripts or entry.type not in acct.locking_scripts:
                continue
            if acct.id not in self.auth_scripts:
                return False
            runtime = tapescript_runtime.get(entry.id, {**tapescript_runtime})
            if 'cache' not in runtime:
                runtime['cache'] = {}
            if 'sigfield1' not in runtime['cache']:
                runtime['cache'] = {
                    **runtime['cache'],
                    **entry.get_sigfields(tapescript_runtime=tapescript_runtime)
                }
            if not acct.validate_script(entry.type, self.auth_scripts[acct.id], runtime):
                return False

        # finally check that correspondent accounting is not violated
        if len(ledgers) > 1:
            accounts = [e.account for e in self.entries]

            for acct in accounts:
                acct: Account
                if acct.type is AccountType.NOSTRO_ASSET and type(acct.details) is str \
                    and acct.details != acct.id and await Identity.find(acct.details):
                    # Nostro account must have equivalent Vostro account
                    nostro = acct
                    await nostro.ledger().reload()
                    cor: Correspondence = await Correspondence.query().contains(
                        'identity_ids', nostro.ledger.identity_id
                    ).contains('identity_ids', nostro.details).first()
                    if cor is None:
                        continue
                    accts = await cor.get_accounts()
                    if acct.details not in accts:
                        return False
                    if AccountType.VOSTRO_LIABILITY not in accts[acct.details]:
                        return False
                    vostro = accts[acct.details][AccountType.VOSTRO_LIABILITY]
                    if vostro.id not in [a.id for a in accounts]:
                        # each nostro Entry must have an offsetting vostro Entry
                        return False
                    offsetting_entry = [e for e in self.entries if e.account_id == vostro.id]
                    if len(offsetting_entry) < 1:
                        return False
                    if offsetting_entry[0].amount != entry.amount:
                        return False

                if acct.type is AccountType.VOSTRO_LIABILITY and type(acct.details) is str \
                    and acct.details != acct.id and await Identity.find(acct.details):
                    # Vostro account must have equivalent Nostro account
                    vostro = acct
                    await vostro.ledger().reload()
                    cor: Correspondence = await Correspondence.query().contains(
                        'identity_ids', vostro.ledger.identity_id
                    ).contains('identity_ids', vostro.details).first()
                    if cor is None:
                        continue
                    accts = await cor.get_accounts()
                    if acct.details not in accts:
                        return False
                    if AccountType.NOSTRO_ASSET not in accts[acct.details]:
                        return False
                    nostro = accts[acct.details][AccountType.NOSTRO_ASSET]
                    if nostro.id not in [a.id for a in accounts]:
                        # each nostro Entry must have an offsetting nostro Entry
                        return False
                    offsetting_entry = [e for e in self.entries if e.account_id == nostro.id]
                    if len(offsetting_entry) < 1:
                        return False
                    if offsetting_entry[0].amount != entry.amount:
                        return False

        return True

    async def save(self, tapescript_runtime: dict = {}, reload: bool = False) -> Transaction:
        """Validate the transaction, save the entries, then save the
            transaction.
        """
        assert await self.validate(tapescript_runtime, reload), 'cannot save an invalid Transaction'
        for e in self.entries:
            await e.save()
        return await super().save()

    async def archive(self) -> ArchivedTransaction:
        """Archive the Transaction. If it has already been archived,
            return the existing ArchivedTransaction.
        """
        archived_txn_id = ArchivedTransaction.generate_id({**self.data})
        try:
            return await ArchivedTransaction.insert({**self.data})
        except Exception as e:
            return await ArchivedTransaction.find(archived_txn_id)
